/*
 * Basic four-word fraction declaration and manipulation.
 *
 * When adding quadword support for 32 bit machines, we need
 * to be a little careful as double multiply uses some of these
 * macros: (in op-2.h)
 * _FP_MUL_MEAT_2_wide() uses _FP_FRAC_DECL_4, _FP_FRAC_WORD_4,
 * _FP_FRAC_ADD_4, _FP_FRAC_SRS_4
 * _FP_MUL_MEAT_2_gmp() uses _FP_FRAC_SRS_4 (and should use
 * _FP_FRAC_DECL_4: it appears to be broken and is not used 
 * anywhere anyway. )
 *
 * I've now fixed all the macros that were here from the sparc64 code.
 * [*none* of the shift macros were correct!] -- PMM 02/1998
 * 
 * The only quadword stuff that remains to be coded is: 
 * 1) the conversion to/from ints, which requires 
 * that we check (in op-common.h) that the following do the right thing
 * for quadwords: _FP_TO_INT(Q,4,r,X,rsz,rsg), _FP_FROM_INT(Q,4,X,r,rs,rt)
 * 2) multiply, divide and sqrt, which require:
 * _FP_MUL_MEAT_4_*(R,X,Y), _FP_DIV_MEAT_4_*(R,X,Y), _FP_SQRT_MEAT_4(R,X,q),
 * This also needs _FP_MUL_MEAT_Q and _FP_DIV_MEAT_Q to be defined to
 * some suitable _FP_MUL_MEAT_4_* macros in sfp-machine.h.
 * [we're free to choose whatever FP_MUL_MEAT_4_* macros we need for
 * these; they are used nowhere else. ]
 */


/* FP variables are really pointers to structs: */
typedef struct 
{
   _FP_COMMON_DECL;                               /* from op-common-pre.h */
   _FP_W_TYPE f[4];
} _FP_4_struct;

typedef _FP_4_struct* _FP_4;

INLINE void _FP_FRAC_COPY_4(_FP_4 D, _FP_4 S)
{
   D->f[0] = S->f[0];
   D->f[1] = S->f[1];
   D->f[2] = S->f[2];
   D->f[3] = S->f[3];
}

INLINE _FP_W_TYPE* _FP_FRAC_HIGH_4(_FP_4 X)
{
   return &(X->f[3]);
}

INLINE _FP_W_TYPE* _FP_FRAC_LOW_4(_FP_4 X)
{
   return &(X->f[0]);
}

INLINE _FP_W_TYPE* _FP_FRAC_WORD_4(_FP_4 X, _FP_I_TYPE w)
{
   return &(X->f[w]);
}

INLINE void _FP_FRAC_SET_ZERO_4(_FP_4 X)
{
   X->f[0] = X->f[1] = X->f[2] = X->f[3] = 0;
}

INLINE void _FP_FRAC_SET_MIN_4(_FP_4 X)
{
   X->f[1] = X->f[2] = X->f[3] = 0;
   X->f[0] = 1;
}

INLINE void _FP_FRAC_SET_NAN_4(char fs, _FP_4 X)
{
   X->f[0] = _FP_NANFRAC(fs,0);
   X->f[1] = _FP_NANFRAC(fs,1);
   X->f[2] = _FP_NANFRAC(fs,2);
   X->f[3] = _FP_NANFRAC(fs,3);
}

INLINE void _FP_FRAC_SLL_4(_FP_4 X, _FP_I_TYPE N)
{
   _FP_I_TYPE up, down, skip, i;
   skip = N / _FP_W_TYPE_SIZE;
   up = N % _FP_W_TYPE_SIZE;
   down = _FP_W_TYPE_SIZE - up;
   for (i = 3; i > skip; i--)
      X->f[i] = X->f[i-skip] << up | X->f[i-skip-1] >> down;
   X->f[i] = X->f[0] << up;
   for (--i; i >= 0; i--)
      X->f[i] = 0;
}

INLINE void _FP_FRAC_SRL_4(_FP_4 X, _FP_I_TYPE N)
{
   _FP_I_TYPE up, down, skip, i;
   skip = N / _FP_W_TYPE_SIZE;
   down = N % _FP_W_TYPE_SIZE;
   up = _FP_W_TYPE_SIZE - down;
   for (i = 0; i < 3-skip; i++)
      X->f[i] = X->f[i+skip] >> down | X->f[i+skip+1] << up;
   X->f[i] = X->f[3] >> down;
   for (++i; i < 4; i++)
      X->f[i] = 0;
}

/* Right shift with sticky-lsb. 
 * What this actually means is that we do a standard right-shift,
 * but that if any of the bits that fall off the right hand side
 * were one then we always set the LSbit.
 */
INLINE void _FP_FRAC_SRS_4(_FP_4 X, _FP_I_TYPE N, _FP_I_TYPE size)
{
   _FP_I_TYPE up, down, skip, i;
   _FP_W_TYPE s;
   skip = N / _FP_W_TYPE_SIZE;
   down = N % _FP_W_TYPE_SIZE;
   up = _FP_W_TYPE_SIZE - down;
   
   for (s = i = 0; i < skip; i++)
      s |= X->f[i];
   s |= X->f[i] << up;
   /* s is now != 0 if we want to set the LSbit */
   for (i = 0; i < 3-skip; i++)
      X->f[i] = X->f[i+skip] >> down | X->f[i+skip+1] << up;
   X->f[i] = X->f[3] >> down;
   for (++i; i < 4; i++)
      X->f[i] = 0;
   /* don't fix the LSbit until the very end when we're sure f[0] is stable */
   X->f[0] |= (s != 0);
}

INLINE void __FP_FRAC_ADD_4(_FP_W_TYPE r[], 
                            _FP_W_TYPE x3, _FP_W_TYPE x2, _FP_W_TYPE x1, _FP_W_TYPE x0,
                            _FP_W_TYPE y3, _FP_W_TYPE y2, _FP_W_TYPE y1, _FP_W_TYPE y0)
{
   r[0] = x0 + y0;
   r[1] = x1 + y1 + (r[0] < x0);
   r[2] = x2 + y2 + (r[1] < x1);
   r[3] = x3 + y3 + (r[2] < x2);
}

INLINE void __FP_FRAC_SUB_4(_FP_W_TYPE r[], 
                            _FP_W_TYPE x3, _FP_W_TYPE x2, _FP_W_TYPE x1, _FP_W_TYPE x0,
                            _FP_W_TYPE y3, _FP_W_TYPE y2, _FP_W_TYPE y1, _FP_W_TYPE y0)
{
   r[0] = x0 - y0;
   r[1] = x1 - y1 - (r[0] > x0);
   r[2] = x2 - y2 - (r[1] > x1);
   r[3] = x3 - y3 - (r[2] > x2);
}

INLINE void _FP_FRAC_ADD_4(_FP_4 R, _FP_4 X, _FP_4 Y)
{
   __FP_FRAC_ADD_4(R->f,
                   X->f[3],X->f[2],X->f[1],X->f[0],
                   Y->f[3],Y->f[2],Y->f[1],Y->f[0]);
}

INLINE void _FP_FRAC_SUB_4(_FP_4 R, _FP_4 X, _FP_4 Y)
{
   __FP_FRAC_SUB_4(R->f,
                   X->f[3],X->f[2],X->f[1],X->f[0],
                   Y->f[3],Y->f[2],Y->f[1],Y->f[0]);
}

INLINE void _FP_FRAC_ADDI_4(_FP_4 X, _FP_W_TYPE I)
{
   /* I always wanted to be a lisp programmer :-> */
   X->f[3] += ((X->f[2] += ((X->f[1] += ((X->f[0] += I) < X->f[0])) < X->f[1]) < X->f[2]));
}

INLINE _FP_I_TYPE _FP_FRAC_ZEROP_4(_FP_4 X)
{
   return ((X->f[0] | X->f[1] | X->f[2] | X->f[3]) == 0);
}

INLINE _FP_I_TYPE _FP_FRAC_NEGP_4(_FP_4 X)
{
   return ((_FP_WS_TYPE)X->f[3] < 0);
}

INLINE _FP_I_TYPE _FP_FRAC_OVERP_4(char fs, _FP_4 X)
{
   return (X->f[3] & _FP_OVERFLOW(fs));
}

INLINE _FP_I_TYPE _FP_FRAC_EQ_4(_FP_4 X, _FP_4 Y)
{
   return (X->f[0] == Y->f[0] && X->f[1] == Y->f[1] && X->f[2] == Y->f[2] && X->f[3] == Y->f[3]);
}

INLINE _FP_I_TYPE _FP_FRAC_GT_4(_FP_4 X, _FP_4 Y)
{
   return (X->f[3] > Y->f[3]
           || (X->f[3] == Y->f[3] && (X->f[2] > Y->f[2]
                                      || (X->f[2] == Y->f[2] && (X->f[1] > Y->f[1]
                                                                || (X->f[1] == Y->f[1]
                                                                    && X->f[0] > Y->f[0])
                                         ) 
                                         )
              )
              )
      );
}

INLINE _FP_I_TYPE _FP_FRAC_GE_4(_FP_4 X, _FP_4 Y)
{
   return (X->f[3] > Y->f[3]
           || (X->f[3] == Y->f[3] && (X->f[2] > Y->f[2]
                                      || (X->f[2] == Y->f[2] && (X->f[1] > Y->f[1]
                                                                || (X->f[1] == Y->f[1]
                                                                    && X->f[0] >= Y->f[0])
                                         )
                                         )
              )
              )
      );
}

INLINE _FP_I_TYPE _FP_FRAC_CLZ_4(_FP_4 X)
{
   if (X->f[3])
      return __FP_CLZ(X->f[3]);
   else if (X->f[2])
      return _FP_W_TYPE_SIZE + __FP_CLZ(X->f[2]);
   else if (X->f[1])
      return _FP_W_TYPE_SIZE*2 + __FP_CLZ(X->f[1]);
   else
      return _FP_W_TYPE_SIZE*3 + __FP_CLZ(X->f[0]);
}

INLINE void _FP_UNPACK_RAW_4(char fs, _FP_4 X, void *val)
{
   if (fs == 's')
   {
#ifdef __FP_S_IS_4_WORDS
      union _FP_UNION_S *flo = (union _FP_UNION_S *) val;
      X->f[0] = flo->bits.frac0;
      X->f[1] = flo->bits.frac1;
      X->f[2] = flo->bits.frac2;
      X->f[3] = flo->bits.frac3;
      X->e    = flo->bits.exp;
      X->s    = flo->bits.sign;
#endif
   }
   else if (fs == 'd')
   {
#ifdef __FP_D_IS_4_WORDS
      union _FP_UNION_D *flo = (union _FP_UNION_D *) val;
      X->f[0] = flo->bits.frac0;
      X->f[1] = flo->bits.frac1;
      X->f[2] = flo->bits.frac2;
      X->f[3] = flo->bits.frac3;
      X->e    = flo->bits.exp;
      X->s    = flo->bits.sign;
#endif
   }
   else if (fs == 'q')
   {
#ifdef __FP_Q_IS_4_WORDS
      union _FP_UNION_Q *flo = (union _FP_UNION_Q *) val;
      X->f[0] = flo->bits.frac0;
      X->f[1] = flo->bits.frac1;
      X->f[2] = flo->bits.frac2;
      X->f[3] = flo->bits.frac3;
      X->e    = flo->bits.exp;
      X->s    = flo->bits.sign;
#endif
   }
}


INLINE void _FP_PACK_RAW_4(char fs, void *val, _FP_4 X)
{
   if (fs == 's')
   {
#ifdef __FP_S_IS_4_WORDS
      union _FP_UNION_S *flo = (union _FP_UNION_S *) val;
      flo->bits.frac0 = X->f[0];
      flo->bits.frac1 = X->f[1];
      flo->bits.frac2 = X->f[2];
      flo->bits.frac3 = X->f[3];
      flo->bits.exp   = X->e;
      flo->bits.sign  = X->s;
#endif
   }
   else if (fs == 'd')
   {
#ifdef __FP_D_IS_4_WORDS
      union _FP_UNION_D *flo = (union _FP_UNION_D *) val;
      flo->bits.frac0 = X->f[0];
      flo->bits.frac1 = X->f[1];
      flo->bits.frac2 = X->f[2];
      flo->bits.frac3 = X->f[3];
      flo->bits.exp   = X->e;
      flo->bits.sign  = X->s;
#endif
   }
   else
   {
#ifdef __FP_Q_IS_4_WORDS
      union _FP_UNION_Q *flo = (union _FP_UNION_Q *) val;
      flo->bits.frac0 = X->f[0];
      flo->bits.frac1 = X->f[1];
      flo->bits.frac2 = X->f[2];
      flo->bits.frac3 = X->f[3];
      flo->bits.exp   = X->e;
      flo->bits.sign  = X->s;
#endif
   }
}


/*
 * Internals 
 */
/* Multiply the two normalised fractional parts of X and Y together
 * and put the result in R, normalised (adjusting R->e appropriately).
 * We already know that both X and Y are normalized non-zero numbers.
 * We use an algorithm borrowed from the NetBSD kernel.
 * It doesn't use any asm() assists; feel free to write one that does...
 *
 * The fraction of the product is built in the usual stepwise fashion.
 * Each step consists of shifting the accumulator right one bit
 * (maintaining any guard bits) and, if the next bit in y is set,
 * adding the multiplicand (x) to the accumulator.  Then, in any case,
 * we advance one bit leftward in y.  Algorithmically:
 *
 *      R = 0;
 *      for (bit = 0; bit < FP_NMANT; bit++) {
 *              sticky |= R & 1, A >>= 1;
 *              if (R & (1 << bit))
 *                      R += X;
 *      }
 *
 * (X and Y here represent the mantissas of x and y respectively.)
 * The resultant accumulator (R) is the product's mantissa.  It may
 * be as large as 11.11111... in binary and hence may need to be
 * shifted right, but at most one bit. This single bit shift is 
 * done if necessary by the generic code in op-common-post.h.
 *
 * In the algorithm above, the bits in y are inspected one at a time.
 * We will pick them up 32 at a time and then deal with those 32, one
 * at a time. 
 */
/* some helper functions for the multiply: */
INLINE void __FP_FRAC_SRS32_4(_FP_4 X)
{
   /* Fast 32 bit shift right with sticky bit, equivalent to _FP_FRAC_SRS_4(X,32,n) */
   X->f[0] = X->f[1] | (X->f[0] != 0);
   X->f[1] = X->f[2];
   X->f[2] = X->f[3];
}

INLINE void __FP_FRAC_SRS1_4(_FP_4 X)
{
   /* ditto, but single bit SRS */
   _FP_I_TYPE r = 1, l = _FP_W_TYPE_SIZE - 1;
   X->f[0] = (X->f[0] >> r) | (X->f[1] << l) | (X->f[0] & 1);
   X->f[1] = (X->f[1] >> r) | (X->f[2] << l);
   X->f[2] = (X->f[2] >> r) | (X->f[3] << l);
   X->f[3] = (X->f[3] >> r);
}

/* do a single step of the multiply */
INLINE void __FP_MUL_STEP_4(_FP_4 R, _FP_4 X, _FP_W_TYPE curw, _FP_W_TYPE maxbit)
{
   _FP_W_TYPE bit = 1;
   if (curw == 0)                                 /* optimisation: if m==0 we'll never */
   {                                              /* add in the loop below, so skip it */
      __FP_FRAC_SRS32_4(R);
      return;
   }
   do 
   {
      __FP_FRAC_SRS1_4(R);
      if (bit & curw) 
         _FP_FRAC_ADD_4(R,R,X);
      bit <<= 1;
   } while (bit <= maxbit);
}

INLINE void _FP_MUL_MEAT_4_slow(char fs, _FP_4 R, _FP_4 X, _FP_4 Y)
{
   _FP_FRAC_SET_ZERO_4(R);
   printk("MUL_4: X = %08lx %08lx %08lx %08lx\n", X->f[3],X->f[2],X->f[1],X->f[0]);
   printk("MUL_4: Y = %08lx %08lx %08lx %08lx\n", Y->f[3],Y->f[2],Y->f[1],Y->f[0]);
   __FP_MUL_STEP_4(R,X,Y->f[0], (_FP_W_TYPE)1 << (_FP_W_TYPE_SIZE-1));
   __FP_MUL_STEP_4(R,X,Y->f[1], (_FP_W_TYPE)1 << (_FP_W_TYPE_SIZE-1));
   __FP_MUL_STEP_4(R,X,Y->f[2], (_FP_W_TYPE)1 << (_FP_W_TYPE_SIZE-1));
   /* In the final step, we can stop shifting as soon as we've got past
    * the highest set bit in Y->f[3].
    */
   __FP_MUL_STEP_4(R,X,Y->f[3], Y->f[3]);
   printk("MUL_4: R = %08lx %08lx %08lx %08lx\n", R->f[3],R->f[2],R->f[1],R->f[0]);
}

/* Another inefficient C implementation: this one is quadword division */
INLINE void _FP_DIV_MEAT_4_slow(char fs, _FP_4 R, _FP_4 X, _FP_4 Y)
{
   /* This algorithm is the one used in the NetBSD kernel,
    * although it is a hideously inefficient version.
    * The common code has already dealt with special cases and
    * set the result exponent to X->e - Y->e.
    * R = (X/Y) * 2^{Xexp - Yexp} and X,Y \in [1.0,2.0)
    * so the quotient's mantissa X/Y will be \in [0.5,2.0).
    * Moreover, it will be <1.0 iff X<Y, in which case we must shift
    * it left one bit, adjusting the exponent, to normalise the result. 
    *
    * Feel free to implement a better routine; it is possible
    * to cut out a lot of the multi-word arithmetic used here:
    * since we work down through the result we can deal with 
    * each word separately.
    */
   _FP_DECL(_FP_4,RDR);                           /* remainder */
   _FP_DECL(_FP_4,BIT);
   BIT->f[0] = BIT->f[1] = BIT->f[2] = 0;
   BIT->f[3] = _FP_IMPLBIT(fs);
   _FP_FRAC_SET_ZERO_4(R);
   _FP_FRAC_COPY_4(RDR,X);
   
   if (_FP_FRAC_GT_4(Y,X))                        /* will need to adjust */
   {
      _FP_FRAC_SLL_4(RDR,1);
      R->e--;
   }
   for(; !_FP_FRAC_ZEROP_4(BIT); _FP_FRAC_SRL_4(BIT,1), _FP_FRAC_SLL_4(RDR,1))
   {
      if (_FP_FRAC_GE_4(RDR,Y))
      {
         _FP_FRAC_ADD_4(R,R,BIT);
         _FP_FRAC_SUB_4(RDR,RDR,Y);
      }
   }
   /* If the rdr is non-zero we probably ought to set
    * the sticky bit... 
    */
}

/* Sqrt routine. Note that we must define and initialise S and T
 * here. The algorithm is partly done in op-common-post.h and
 * partly here (see the notes in that file). Basically, we
 * have to do a fixed-point square root. We have one extra bit
 * of precision [ie, X's fractional part was shifted left 1
 * before we were called and we should return R such that 
 * shifting it right one normalises it.] 
 * The algorithm here is the following:
 * R = S = 0;
 * for (q = 2^MAXBITS; q > 0; q >>=1)
 * {
 *    T = S + q;
 *    if (T <= X)
 *    {
 *       S = T + q;
 *       X -= T;
 *       R += q;
 *    }
 *    X *= 2;
 * }
 */
INLINE void _FP_SQRT_MEAT_4(_FP_4 R, _FP_4 X, _FP_W_TYPE q)
{
   _FP_DECL(_FP_4,S);
   _FP_DECL(_FP_4,T);
   _FP_DECL(_FP_4,Q);
   
   /* This is a very inefficient implementation:
    * the obvious optimisation is to handle each word separately
    * rather than using quadword arithmetic all the time.
    * This is left as an exercise for the reader...
    */
   _FP_FRAC_SET_ZERO_4(S);

   printk("SQRT_4: X frac = %08lx %08lx %08lx %08lx\n",X->f[3],X->f[2],X->f[1],X->f[0]);

   for(Q->f[3] = q, Q->f[2] = Q->f[1] = Q->f[0] = 0; !_FP_FRAC_ZEROP_4(Q) ;_FP_FRAC_SRL_4(Q,1))
   {
      _FP_FRAC_ADD_4(T,S,Q);
      if(_FP_FRAC_GE_4(X,T))
      {
         _FP_FRAC_ADD_4(S,T,Q);
         _FP_FRAC_SUB_4(X,X,T);
         _FP_FRAC_ADD_4(R,R,Q);
      }
      _FP_FRAC_SLL_4(X,1);
   }
   printk("result frac = %08lx %08lx %08lx %08lx\n",R->f[3],R->f[2],R->f[1],R->f[0]);
}


/* Assembly/disassembly for converting to/from integral types.  
 * No shifting or overflow handled here.
 */
/* Put the FP value X into r, which is an integer of size rsize. */
INLINE unsigned _FP_CONV_I_TYPE _FP_FRAC_ASSEMBLE_4(_FP_4 X, _FP_I_TYPE rsize)
{
   unsigned _FP_CONV_I_TYPE r;
   
   if (rsize <= _FP_W_TYPE_SIZE)
      r = X->f[0];
   else if (rsize <= 2*_FP_W_TYPE_SIZE)
#if _FP_CONV_I_TYPE_SIZE > _FP_W_TYPE_SIZE
      r = X->f[0] + (((_FP_CONV_I_TYPE)X->f[1]) << _FP_W_TYPE_SIZE);
#else
      /* can't happen unless caller is buggy */
      printk("_FP_FRAC_ASSEMBLE_4: bogus rsize (bug!)\n");
#endif
   else
   {
#if _FP_CONV_I_TYPE_SIZE > 2*_FP_W_TYPE_SIZE
      /* I'm feeling lazy so we deal with int == 3words (implausible)
       * and int == 4words as a single case.                         
       */
      r = ((_FP_CONV_I_TYPE)X->f[3]) << _FP_W_TYPE_SIZE;
      r += X->f[2];
      r <<= _FP_W_TYPE_SIZE;
      r += X->f[1];
      r <<= _FP_W_TYPE_SIZE;
      r += X->f[0];
#else
      /* can't happen unless caller is buggy */
      printk("_FP_FRAC_ASSEMBLE_4: bogus rsize (bug!)\n");
#endif
   }
   return r;
}


/* "No disassemble Number Five!" */
/* move an integer of size rsize into X's fractional part. We rely on
 * the _f[] array consisting of words of size _FP_W_TYPE_SIZE to avoid
 * having to mask the values we store into it.
 * The interwoven if() and #if structure looks ugly but is perfectly logical :->
 */
INLINE void _FP_FRAC_DISASSEMBLE_4(_FP_4 X, unsigned _FP_CONV_I_TYPE r, _FP_I_TYPE rsize)
{
   X->f[0] = r;
   if (rsize <= _FP_W_TYPE_SIZE)
   {
      X->f[1] = 0;
   }
   else
   {
#if _FP_CONV_I_TYPE_SIZE > _FP_W_TYPE_SIZE
      X->f[1] = r >> _FP_W_TYPE_SIZE;
      if (rsize <= 2 * _FP_W_TYPE_SIZE)
      {
         X->f[2] = 0;
      }
      else
      {
#if _FP_CONV_I_TYPE_SIZE > 2*_FP_W_TYPE_SIZE
         X->f[2] = r >> 2 * _FP_W_TYPE_SIZE;
         if (rsize <= 3 * _FP_W_TYPE_SIZE)
         {
            X->f[3] = 0;
         }
         else
         {
#if _FP_CONV_I_TYPE_SIZE > 3*_FP_W_TYPE_SIZE
            X->f[3] = r >> 3 * _FP_W_TYPE_SIZE;
#else
            /* can't happen unless caller is buggy */
            printk("_FP_FRAC_DISASSEMBLE_4: bogus rsize (bug!)\n");
#endif 
         }
#else
      /* can't happen unless caller is buggy */
      printk("_FP_FRAC_DISASSEMBLE_4: bogus rsize (bug!)\n");
#endif    
      }
#else
      /* can't happen unless caller is buggy */
      printk("_FP_FRAC_DISASSEMBLE_4: bogus rsize (bug!)\n");
#endif    
   }
}

/* The only conversion we need to put in this file is a 4->4 conversion.
 * This is non-trivial because the number of working bits used may be 
 * different for the types we are converting between.
 */
INLINE void _FP_FRAC_CONV_4_4(char dfs, char sfs, _FP_4 D, _FP_4 S)
{
   D->f[0] = S->f[0]; D->f[1] = S->f[1]; D->f[2] = S->f[2]; D->f[3] = S->f[3];
   if (_FP_WFRACBITS(sfs) > _FP_WFRACBITS(dfs))
      _FP_FRAC_SRS_4(D, _FP_WFRACBITS(sfs) - _FP_WFRACBITS(dfs), _FP_WFRACBITS(sfs)); 
   else if (_FP_WFRACBITS(sfs) < _FP_WFRACBITS(dfs))
      _FP_FRAC_SLL_4(D, _FP_WFRACBITS(dfs) - _FP_WFRACBITS(sfs));
}
