/* soft-fp header file: set up the FP emulation system.
 * Must be #included before single.h/double.h/quad.h.
 *
 * The exact ordering of #includes in this file is a little tricky because 
 * we use inline functions for efficiency. This means that we have to order
 * things so that all functions are defined before they are used. 
 * The structure is as follows:
 * sfp-machine.h : machine dependent stuff. This has to come first
 * because it defines things like _FP_W_TYPE_SIZE which are fundamental.
 * single-pre.h, double-pre.h, quad-pre.h : those defns which are required
 * by the rest of the system, such as _FP_FRACTBITS_D
 * op-common-pre.h : parts of the common code which op-n.h depend upon
 * op-dispatch-pre.h : functions/macros dispatching according to fs/n
 * op-4.h : the real meat for dealing with n-word ops. The order
 * op-2.h : here is important
 * op-1.h
 * op-dispatch.h : more functions/macros dispatching according to fs/n
 * op-common-post.h : common code dependent on op-n.h
 * rest of single/double/quad.h : functions the user actually uses
 */
#ifndef SOFT_FP_H
#define SOFT_FP_H

#include <linux/kernel.h>                         /* for printk() */

#include "sfp-machine.h"                          /* machine dependencies */

/* these are pretty important so we define them early */
#define _FP_WORKBITS 3
#define _FP_WORK_LSB    ((_FP_W_TYPE)1 << 3)
#define _FP_WORK_ROUND  ((_FP_W_TYPE)1 << 2)
#define _FP_WORK_GUARD  ((_FP_W_TYPE)1 << 1)
#define _FP_WORK_STICKY ((_FP_W_TYPE)1 << 0)
/* shift right 0, how useful */

#define FP_RND_NEAREST 0
#define FP_RND_ZERO 1
#define FP_RND_PINF 2
#define FP_RND_MINF 3
#define FP_ROUNDMODE FP_RND_NEAREST

/* Sigh.  Silly things longlong.h needs.  */
#define UWtype		_FP_W_TYPE
#define W_TYPE_SIZE	_FP_W_TYPE_SIZE

typedef int SItype __attribute__((mode(SI)));
typedef int DItype __attribute__((mode(DI)));
typedef unsigned int USItype __attribute__((mode(SI)));
typedef unsigned int UDItype __attribute__((mode(DI)));
#if _FP_W_TYPE_SIZE == 32
typedef unsigned int UHWtype __attribute__((mode(HI)));
#elif _FP_W_TYPE_SIZE == 64
typedef USItype UHWtype;
#endif

/* possible values for the condition (c) field of FP structs */
#define FP_CLS_NORMAL 0
#define FP_CLS_ZERO 1
#define FP_CLS_INF 2
#define FP_CLS_NAN 3

/* This had better stay a macro so we can use it in switch() statements */
#define _FP_CLS_COMBINE(x,y)	(((x) << 2) | (y))

/* on with the #includes */

#include "single-pre.h"
#include "double-pre.h"
#include "quad-pre.h"

#include "op-common-pre.h"
#include "op-dispatch-pre.h"
#include "op-4.h"
#include "op-2.h"
#include "op-1.h"
#include "op-dispatch.h"
#include "op-common-post.h"


#endif /* not SOFT_FP_H */
