/*
 * Program to test the FP emulation
 * Copyright (C) 03/1998 Peter Maydell (pmaydell@chiark.greenend.org.uk)
 * This program is GPLed.
 *
 * TESTING STRATEGY
 *
 * We need to do some serious testing of every emulated opcode.
 * This is going to be a real pain :->
 * Approach: have a simple program to read a config file
 * and generate appropriate assembly and then test it.
 * Config file looks like
 * FMULD 0 0 0 02312313    0 0 0 0xffff0000    0 0 0 0x39383883
 * (opcode) (first operand) (2nd operand)  (result)
 * The gaps can be any whitespace.
 * Note that parameters are always 16 bytes even for single and double ops.
 * Each op is represented using four 32bit values, MSbyte first. 
 * Unrequired significant words are dropped. 
 * The return value of the comparison ops is the condition code 
 * field of the fsr; the fsr is masked appropriately and returned
 * in the fourth word of the result field.
 * We calculate the result field by running a similar program under SunOS.
 *
 * Note that we don't attempt to test all possible register 
 * combinations for an opcode.
 *
 * Usage:
 * tester -c : takes a config file with or without results on stdin and generates 
 * one with a results column on stdout.
 *
 * tester : takes a config file with results on stdin and 
 * compares the results generated with the ones in the config file.
 * Discrepancies are printed to stdout.
 *
 * NB: requires GCC, as we use its 'long long' extension. This isn't 
 * a huge problem, especially since we need gcc to compile the kernel
 * FP emulation code in the first place...
 * This restriction doesn't apply if your long is 64 bits anyway;
 * just define INT64 appropriately below.
 */


/* #define SPARCV9  
 * If compiling for V9, include code to test V9 specific insns.
 * Even though we emulate them, we can't test them because gas won't
 * let us :-< 
 */

/* You'll need at least a 32bit type and a 64bit type; this is 
 * the SPARC32 set of #defines. 
 * It is assumed that passing a UINT32 or a UINT64 to or from 
 * an asm() expression with a constraint string of "f" or "=f"
 * will cause it to be correctly loaded into a 32bit FP register
 * or a 64bit FP register (ie an aligned pair of 32bit registers). 
 * You also need format strings for printf and scanf, sigh.
 * UINT32SCANFFORMAT should be one to read a hex UINT32, eg "%x" or "%lx".
 * UINT32PRINTFFORMAT should be one to output a hex UINT32, eg "%08lx".
 * The quotes are required.
 */
#define INT64 long long
#define INT32 long
#define UINT64 unsigned INT64
#define UINT32 unsigned INT32
#define UINT32SCANFFORMAT "%lx"
#define UINT32PRINTFFORMAT "%08lx"

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>                                /* for isspace() */
#include <sys/utsname.h>                          /* for uname() */
#include <time.h>                                 /* for asctime(), localtime(), time() */

ssize_t getline(char **, size_t *, FILE*);

/* Possible modes */
#define COMPARE 0
#define GENERATE 1
/* type of the fpop functions. We use this macro rather than a typedef
 * for hysterical raisins.
 */
#define FLOATFUNC(x) void x(UINT32 op1[], UINT32 op2[], UINT32 res[])

int extract(char *line, FLOATFUNC((**opcode)), UINT32 op1[], 
            UINT32 op2[], UINT32 res[]);
void usage(void);

/* Declare all the functions which correspond to tested insns */
/* quadword instructions (19) */
FLOATFUNC(FMOVQ);
FLOATFUNC(FNEGQ);
FLOATFUNC(FABSQ);
FLOATFUNC(FSQRTQ);
FLOATFUNC(FADDQ);
FLOATFUNC(FSUBQ);
FLOATFUNC(FMULQ); 
FLOATFUNC(FDIVQ); 
FLOATFUNC(FDMULQ);
FLOATFUNC(FQTOX);
FLOATFUNC(FXTOQ);
FLOATFUNC(FQTOS);
FLOATFUNC(FQTOD);
FLOATFUNC(FITOQ); 
FLOATFUNC(FSTOQ);
FLOATFUNC(FDTOQ);
FLOATFUNC(FQTOI);
FLOATFUNC(FCMPQ);
FLOATFUNC(FCMPEQ);
/* single/double instructions (subnormal) (17) */
FLOATFUNC(FSQRTS);
FLOATFUNC(FSQRTD);
FLOATFUNC(FADDS);
FLOATFUNC(FADDD);
FLOATFUNC(FSUBS);
FLOATFUNC(FSUBD);
FLOATFUNC(FMULS);
FLOATFUNC(FMULD);
FLOATFUNC(FDIVS);
FLOATFUNC(FDIVD);
FLOATFUNC(FSMULD);
FLOATFUNC(FSTOX);
FLOATFUNC(FDTOX);
FLOATFUNC(FDTOS);
FLOATFUNC(FSTOD);
FLOATFUNC(FSTOI);
FLOATFUNC(FDTOI);
/* other stuff V9 doesn't need but V8 might? (14) */
FLOATFUNC(FABSS);
FLOATFUNC(FABSD);
FLOATFUNC(FCMPS);
FLOATFUNC(FCMPES);
FLOATFUNC(FCMPD);
FLOATFUNC(FCMPED);
FLOATFUNC(FMOVS);
FLOATFUNC(FMOVD);
FLOATFUNC(FNEGS);
FLOATFUNC(FNEGD);
FLOATFUNC(FITOS); 
FLOATFUNC(FITOD);
FLOATFUNC(FXTOS);
FLOATFUNC(FXTOD);

/* Now we have arrays so we can map opcode name strings to functions */
#define NUMOPS (19+17+14)
/* Yes, this is ugly. There ought to be some way to do the 
 * function declaration and both arrays in one go.
 */
char *opname[] = 
{
   "FMOVQ", "FNEGQ", "FABSQ", "FSQRTQ", "FADDQ", "FSUBQ", "FMULQ",
   "FDIVQ", "FDMULQ", "FQTOX", "FXTOQ", "FQTOS", "FQTOD", "FITOQ",
   "FSTOQ", "FDTOQ", "FQTOI", "FCMPQ", "FCMPEQ", "FSQRTS", "FSQRTD",
   "FADDS", "FADDD", "FSUBS", "FSUBD", "FMULS", "FMULD", "FDIVS",
   "FDIVD", "FSMULD", "FSTOX", "FDTOX", "FDTOS", "FSTOD", "FSTOI",
   "FDTOI", "FABSS", "FABSD", "FCMPS", "FCMPES", "FCMPD", "FCMPED",
   "FMOVS", "FMOVD", "FNEGS", "FNEGD", "FITOS", "FITOD", "FXTOS", "FXTOD"
};

/* opfn is an array of functions of type FLOATFUNC */
FLOATFUNC((*opfn[])) = 
{
   FMOVQ, FNEGQ, FABSQ, FSQRTQ, FADDQ, FSUBQ, FMULQ, 
   FDIVQ, FDMULQ, FQTOX, FXTOQ, FQTOS, FQTOD, FITOQ, 
   FSTOQ, FDTOQ, FQTOI, FCMPQ, FCMPEQ, FSQRTS, FSQRTD,
   FADDS, FADDD, FSUBS, FSUBD, FMULS, FMULD, FDIVS,
   FDIVD, FSMULD, FSTOX, FDTOX, FDTOS, FSTOD, FSTOI,
   FDTOI, FABSS, FABSD, FCMPS, FCMPES, FCMPD, FCMPED,
   FMOVS, FMOVD, FNEGS, FNEGD, FITOS, FITOD, FXTOS, FXTOD
};


/* Phew! Finally, here's main()... */

int main(int argc, char *argv[])
{
   int mode;
   UINT32 op1[4], op2[4], res[4], ourres[4];
   FLOATFUNC((*opcode));
   char *line = NULL;
   /* getline() mallocs this, so if using this code in a bigger program
    * you should "if(line) free(line)" where we currently exit().
    */
   int linelen = 0;
   int i;
   
   /* Trivial parsing of command line arguments. */
   if (argc == 1)
      mode = COMPARE;
   else if (argc == 2 && !strcmp(argv[1],"-c"))
      mode = GENERATE;
   else
   {
      usage();
      exit(EXIT_FAILURE);
   }
   
   if (mode == GENERATE)
   {
      /* Prepend a note stating where the results field came from... */
      struct utsname info;
      time_t now = time(NULL);
      uname(&info);
      
      printf("# Run through GENERATE at %s# on %s (%s) running %s %s %s\n\n",
             asctime(localtime(&now)),
             info.nodename, info.machine, info.sysname, info.release, info.version);
   }
   
   while (getline(&line, &linelen, stdin) != -1)
   {
      i = extract(line, &opcode, op1, op2, res);
      /* check that what we got back fits the mode */
      if (i == 1 && mode == COMPARE)
      {
         fprintf(stderr,"Can't compare when input file has no results!\n");
         exit(EXIT_FAILURE);
      }
      else if (i == 0 && mode == GENERATE)
      {
         fprintf(stderr,"Can't generate when input file already contains results!\n");
         exit(EXIT_FAILURE);
      }
      else if (i == 2)                                
      { /* comment or blank : pass through verbatim for GENERATE, ignore for COMPARE */
         if (mode == GENERATE)
            fputs (line,stdout);
         continue;
      }
      
      /* perform the FPop... */
      ourres[0] = ourres[1] = ourres[2] = ourres[3] = 0;
      opcode(op1, op2, ourres);
      
      if (mode == COMPARE)
      {
         /* compare our result with the given */
         if (ourres[0] != res[0] || ourres[1] != res[1]
             || ourres[2] != res[2] || ourres[3] != res[3])
         {
            printf("Oops, discrepancy! On input\n%swe got result "
                   UINT32PRINTFFORMAT " " UINT32PRINTFFORMAT " "
                   UINT32PRINTFFORMAT " " UINT32PRINTFFORMAT
                   "...\n",
                   line, ourres[0], ourres[1], ourres[2], ourres[3]);
         }
      }
      else
      {
         /* generate results field: print line, nuking the trailing \n, append calc'd result */
         line[strlen(line)-1] = '\0';
         printf("%s\t"
                UINT32PRINTFFORMAT " " UINT32PRINTFFORMAT " "
                UINT32PRINTFFORMAT " " UINT32PRINTFFORMAT
                "\n",line,ourres[0],ourres[1],ourres[2],ourres[3]);
      }
   }
   exit (EXIT_SUCCESS);
}

void usage(void)
{
   printf("tester v 1.0, copyright 1998 Peter Maydell (pmaydell@chiark.greenend.org.uk)\n\
\nUsage:\n\
tester -c   expects a config file on stdin and outputs a config file\n\
            to stdout with a results column added.\n\
tester      with no args expects a config file with results on stdin and\n\
            warns of discrepancies between the given results and those\n\
            calculated on this machine.\n");
}

int extract(char *line, FLOATFUNC((**opcode)), UINT32 op1[], 
            UINT32 op2[], UINT32 res[])
{
   /* extract the relevant information from one line of the config file and set
    * opcode, op1, op2, res appropriately. If res is not present, we set it to 0
    * and return 1; if the line is blank or a comment we return 2; otherwise we return 0.
    * Note that opcode is actually a function which does the appropriate calculation.
    * We exit here if the line appears to be garbage.
    */
   int i = -1, retcode = 0;
   char name[11];
   
   /* comments start with '#' as first non whitespace char.
    * We set i to -1 initially. If the constant part of the sscanf template matches
    * it then sets i to the # of chars matched so far (because of the %n).
    * Only if sscanf fails to match its initial part will i remain at -1.
    */
   sscanf(line, " #%n",&i);
   if (i != -1)
      return 2;
   
   /* now check for totally blank lines. I'm sure you should be able to
    * do this with sscanf. Ah well; this is less obfuscated, I suppose.
    */
   for (i = 0; i < strlen(line); i++)
   {
      if (!isspace(line[i]))
         break;
   }
   
   if (i == strlen(line))
      return 2;
   
   
   /* line points to: "OPCODE 0 1 2 3  0 1 2 3  0 1 2 3\n"
    * where the last four numbers are optional and the gaps are arbitrary whitespace.
    */
   i = sscanf(line, "%10s"
              " " UINT32SCANFFORMAT " " UINT32SCANFFORMAT " " UINT32SCANFFORMAT
              " " UINT32SCANFFORMAT " " UINT32SCANFFORMAT " " UINT32SCANFFORMAT
              " " UINT32SCANFFORMAT " " UINT32SCANFFORMAT " " UINT32SCANFFORMAT
              " " UINT32SCANFFORMAT " " UINT32SCANFFORMAT " " UINT32SCANFFORMAT
              , name, 
              &op1[0], &op1[1], &op1[2], &op1[3],
              &op2[0], &op2[1], &op2[2], &op2[3], 
              &res[0], &res[1], &res[2], &res[3]);
   
   if (i == 9)                                    /* no res field */
   {
      res[0] = res[1] = res[2] = res[3] = 0;
      retcode = 1;
   }
   else if (i != 13)                              /* oops, input syntax error */
   {
      fprintf(stderr,"Syntax error in input line:\n%s",line);
      exit(EXIT_FAILURE);
   }
   
   for (i = 0; i < NUMOPS; i++)
   {
      if(!strcmp(opname[i],name))
      {
         *opcode = opfn[i];
         break;
      }
   }
   return retcode;
}

/* utility functions for converting the UINT32[] arrays to and from
 * forms suitable for using with asm().
 */
UINT64 int optodouble(UINT32 op[])
{
   UINT64 a = op[2];
   a <<= 32;
   a += op[3];
   return a;
}

UINT32 optosingle(UINT32 op[])
{
   return op[3];
}

void doubletoop(UINT32 res[], UINT64 r)
{
   res[3] = r & 0xffffffff;
   res[2] = r >> 32;
}

void singletoop(UINT32 res[], UINT32 r)
{
   res[3] = r;
}

/* Quad ops are handled somewhat differently. We need to code the
 * loading of values into the FP regs by hand because there's no
 * way to specify that a value goes into a given register, and 
 * asm() doesn't know about 128bit quad values.
 * So you use these macros. The intervening asm should take operands
 * in %f0 and %f4 and put the result in %f8, like this:
 * asm(QUADPREAMBLE "faddq %%f0, %%f4, %%f8" QUADPOSTAMBLE);
 * We assume op1, op2 and res are the UINT32 arrays.
 */
#define QUADPREAMBLE \
"ldd [%1 + 0x00], %%f0; ldd [%1 + 0x08], %%f2; ldd [%2 + 0x00], %%f4; ldd [%2 + 0x08], %%f6; "

#define QUADREGSUSED "%f0", "%f1", "%f2", "%f3", "%f4", "%f5", \
"%f6", "%f7", "%f8", "%f9", "%f10", "%f11", "memory"

#define QUADPOSTAMBLE "; std %%f8, [%0 + 0x00]; std %%f10, [%0 + 0x08]" \
: /* no outputs */ : "r" (res), "r" (op1), "r" (op2) \
: QUADREGSUSED


#ifndef SPARCV9
void v9inst(char *insn)
{
   /* We don't test V9 insns on v8 (gas won't even let us generate them!) */
   fprintf(stderr,"%s is a SPARC V9 instruction!\n",insn);
}
#endif /* not SPARCV9 */

/* Each of these functions is a simple assembly routine to 
 * perform the op with the specified parameters.
 */
FLOATFUNC(FMOVQ)
{
#ifdef SPARCV9
   asm(QUADPREAMBLE "fmovq %%f0, %%f8" QUADPOSTAMBLE);
#else
   v9inst("FMOVQ");
#endif
}

FLOATFUNC(FNEGQ)
{
#ifdef SPARCV9
   asm(QUADPREAMBLE "fnegq %%f0, %%f8" QUADPOSTAMBLE);
#else
   v9inst("FNEGQ");
#endif
}

FLOATFUNC(FABSQ)
{
#ifdef SPARCV9
   asm(QUADPREAMBLE "fabsq %%f0, %%f8" QUADPOSTAMBLE);
#else
   v9inst("FABSQ");
#endif
}
   
FLOATFUNC(FSQRTQ)
{
   asm(QUADPREAMBLE "fsqrtq %%f0, %%f8" QUADPOSTAMBLE);
}

FLOATFUNC(FADDQ)
{
   asm(QUADPREAMBLE "faddq %%f0, %%f4, %%f8" QUADPOSTAMBLE);
}

FLOATFUNC(FSUBQ) 
{
   asm(QUADPREAMBLE "fsubq %%f0, %%f4, %%f8" QUADPOSTAMBLE);
}

FLOATFUNC(FMULQ) 
{
   asm(QUADPREAMBLE "fmulq %%f0, %%f4, %%f8" QUADPOSTAMBLE);
}

FLOATFUNC(FDIVQ)
{ 
   asm(QUADPREAMBLE "fdivq %%f0, %%f4, %%f8" QUADPOSTAMBLE);
}

FLOATFUNC(FDMULQ)
{
   UINT64 a = optodouble(op1), b = optodouble(op2);
   asm("fdmulq %1,%2,%%f0; std %%f0, [%0 + 0x00]; std %%f2, [%0 + 0x08]"
       : /* no outputs */ : "r" (res), "f" (a), "f" (b) 
       : "%f0", "%f1", "%f2", "%f3");
}

FLOATFUNC(FQTOX)
{
#ifdef SPARCV9
   UINT64 r;
   asm("ldd [%1 + 0x00], %%f0; ldd [%1 + 0x08], %%f2; fqtox %%f0, %0"
       : "=f" (r) : "r" (op1) : "%f0", "%f1", "%f2", "%f3");
   doubletoop(res,r);
#else
   v9inst("FQTOX");
#endif
}

FLOATFUNC(FXTOQ)
{
#ifdef SPARCV9   
   UINT64 a = optodouble(op1);
   asm("fdtoq %1, %%f0; std %%f0, [%0]; std %%f2, [%0 + 0x08]"
       : /* no outputs */ : "r" (res), "f" (a) : "%f0", "%f1", "%f2", "%f3", "memory");
#else
   v9inst("FXTOQ");
#endif
}

FLOATFUNC(FQTOS)
{
   UINT32 r;
   asm("ldd [%1 + 0x00], %%f0; ldd [%1 + 0x08], %%f2; fqtos %%f0, %0"
       : "=f" (r) : "r" (op1) : "%f0", "%f1", "%f2", "%f3");
   singletoop(res,r);
}

FLOATFUNC(FQTOD)
{
   UINT64 r;
   asm("ldd [%1 + 0x00], %%f0; ldd [%1 + 0x08], %%f2; fqtod %%f0, %0"
       : "=f" (r) : "r" (op1) : "%f0", "%f1", "%f2", "%f3");
   doubletoop(res,r);
}

FLOATFUNC(FITOQ)
{
   UINT32 a = optosingle(op1);
   asm("fitoq %1, %%f0; std %%f0, [%0]; std %%f2, [%0 + 0x08]"
       : /* no outputs */ : "r" (res), "f" (a) : "%f0", "%f1", "%f2", "%f3", "memory");
} 

FLOATFUNC(FSTOQ)
{
   UINT32 a = optosingle(op1);
   asm("fstoq %1, %%f0; std %%f0, [%0]; std %%f2, [%0 + 0x08]"
       : /* no outputs */ : "r" (res), "f" (a) : "%f0", "%f1", "%f2", "%f3", "memory");
}

FLOATFUNC(FDTOQ)
{
   UINT64 a = optodouble(op1);
   asm("fdtoq %1, %%f0; std %%f0, [%0]; std %%f2, [%0 + 0x08]"
       : /* no outputs */ : "r" (res), "f" (a) : "%f0", "%f1", "%f2", "%f3", "memory");
}

FLOATFUNC(FQTOI)
{
   UINT32 r;
   asm("ldd [%1 + 0x00], %%f0; ldd [%1 + 0x08], %%f2; fqtoi %%f0, %0"
       : "=f" (r) : "r" (op1) : "%f0", "%f1", "%f2", "%f3");
   singletoop(res,r);
}

FLOATFUNC(FCMPQ)
{
   UINT32 r;
   asm(QUADPREAMBLE "fcmpq %%f0, %%f4; st %%fsr, [%0]"
       : /* no outputs */ : "r" (&r), "r" (op1), "r" (op2) :
       "%f0", "%f1", "%f2", "%f3","%f4", "%f5", "%f6", "%f7", "memory");
   singletoop(res, r & 0xc00);
}

FLOATFUNC(FCMPEQ)
{
   UINT32 r;
   asm(QUADPREAMBLE "fcmpeq %%f0, %%f4; st %%fsr, [%0]"
       : /* no outputs */ : "r" (&r), "r" (op1), "r" (op2) :
       "%f0", "%f1", "%f2", "%f3","%f4", "%f5", "%f6", "%f7", "memory");
   singletoop(res, r & 0xc00);
}

/* single/double instructions (subnormal) (17) */
FLOATFUNC(FSQRTS)
{
   UINT32 a = optosingle(op1), r;
   asm("fsqrts %1,%0" : "=f" (r) : "f" (a));
   singletoop(res,r);
}

FLOATFUNC(FSQRTD)
{
   UINT64 a = optodouble(op1), r;
   asm("fsqrtd %1,%0" : "=f" (r) : "f" (a));
   doubletoop(res,r);
}

FLOATFUNC(FADDS)
{
   UINT32 a = optosingle(op1), b = optosingle(op2), r;
   asm("fadds %1,%2,%0" : "=f" (r) : "f" (a), "f" (b));
   singletoop(res,r);
}

FLOATFUNC(FADDD)
{
   UINT64 int a = optodouble(op1), b = optodouble(op2), r;
   asm("faddd %1,%2,%0" : "=f" (r) : "f" (a), "f" (b));
   doubletoop(res,r);
}

FLOATFUNC(FSUBS)
{   
   UINT32 a = optosingle(op1), b = optosingle(op2), r;
   asm("fsubs %1,%2,%0" : "=f" (r) : "f" (a), "f" (b));
   singletoop(res,r);
}

FLOATFUNC(FSUBD)
{
   UINT64 int a = optodouble(op1), b = optodouble(op2), r;
   asm("faddd %1,%2,%0" : "=f" (r) : "f" (a), "f" (b));
   doubletoop(res,r);
}

FLOATFUNC(FMULS)
{
   UINT32 a = optosingle(op1), b = optosingle(op2), r;
   asm("fmuls %1,%2,%0" : "=f" (r) : "f" (a), "f" (b));
   singletoop(res,r);
}

FLOATFUNC(FMULD)
{
   UINT64 int a = optodouble(op1), b = optodouble(op2), r;
   asm("fmuld %1,%2,%0" : "=f" (r) : "f" (a), "f" (b));
   doubletoop(res,r);
}

FLOATFUNC(FDIVS)
{
   UINT32 a = optosingle(op1), b = optosingle(op2), r;
   asm("fdivs %1,%2,%0" : "=f" (r) : "f" (a), "f" (b));
   singletoop(res,r);
}

FLOATFUNC(FDIVD)
{
   UINT64 int a = optodouble(op1), b = optodouble(op2), r;
   asm("fdivd %1,%2,%0" : "=f" (r) : "f" (a), "f" (b));
   doubletoop(res,r);
}

FLOATFUNC(FSMULD)
{
   UINT32 a = optosingle(op1), b = optosingle(op2);
   UINT64 r;
   asm("fsmuld %1,%2,%0" : "=f" (r) : "f" (a), "f" (b));
   doubletoop(res,r);
}

FLOATFUNC(FSTOX)
{
#ifdef SPARCV9
   UINT32 a = optosingle(op1);
   UINT64 r;
   asm("fstox %1,%0" : "=f" (r) : "f" (a));
   doubletoop(res,r);
#else
   v9inst("FSTOX");
#endif
}

FLOATFUNC(FDTOX)
{
#ifdef SPARCV9
   UINT64 r, a = optodouble(op1);
   asm("fdtox %1,%0" : "=f" (r) : "f" (a));
   doubletoop(res,r);
#else
   v9inst("FDTOX");
#endif
}

FLOATFUNC(FDTOS)
{
   UINT64 a = optodouble(op1);
   UINT32 r;
   asm("fdtos %1,%0" : "=f" (r) : "f" (a));
   singletoop(res,r);
}

FLOATFUNC(FSTOD)
{
   UINT32 a = optosingle(op1);
   UINT64 r;
   asm("fstod %1,%0" : "=f" (r) : "f" (a));
   doubletoop(res,r);
}

FLOATFUNC(FSTOI)
{
   UINT32 a = optosingle(op1), r;
   asm("fstoi %1,%0" : "=f" (r) : "f" (a));
   singletoop(res,r);                             /* single because the result is a 32bit word */
}

FLOATFUNC(FDTOI)
{
   UINT64 a = optodouble(op1);
   UINT32 r;
   asm("fdtoi %1,%0" : "=f" (r) : "f" (a));
   singletoop(res,r);
}

/* other stuff V9 doesn't need but V8 might? (14) */
FLOATFUNC(FABSS)
{  
   UINT32 a = optosingle(op1), r;
   asm("fabss %1,%0" : "=f" (r) : "f" (a));
   singletoop(res,r);
}

FLOATFUNC(FABSD)
{
#ifdef SPARCV9
   UINT64 a = optodouble(op1), r;
   asm("fabsd %1,%0" : "=f" (r) : "f" (a));
   doubletoop(res,r);
#else
   v9inst("FABSD");
#endif
}

FLOATFUNC(FCMPS)
{
   UINT32 a = optosingle(op1), b = optosingle(op2), r;
   asm("fcmps %1,%2; st %%fsr, [%0]"
       : /* no outputs */ : "r" (&r), "f" (a), "f" (b) : "memory");
   singletoop(res, r & 0xc00);
}

FLOATFUNC(FCMPES)
{
   UINT32 a = optosingle(op1), b = optosingle(op2), r;
   asm("fcmpes %1,%2; st %%fsr, [%0]"
       : /* no outputs */ : "r" (&r), "f" (a), "f" (b) : "memory");
   singletoop(res, r & 0xc00);
}

FLOATFUNC(FCMPD)
{
   UINT64 a = optodouble(op1), b = optodouble(op2);
   UINT32 r;
   asm("fcmpd %1,%2; st %%fsr, [%0]"
       : /* no outputs */ : "r" (&r), "f" (a), "f" (b) : "memory");
   singletoop(res, r & 0xc00);
}

FLOATFUNC(FCMPED)
{
   UINT64 a = optodouble(op1), b = optodouble(op2);
   UINT32 r;
   asm("fcmped %1,%2; st %%fsr, [%0]"
       : /* no outputs */ : "r" (&r), "f" (a), "f" (b) : "memory");
   singletoop(res, r & 0xc00);
}

FLOATFUNC(FMOVS)
{
   UINT32 a = optosingle(op1),r;
   asm("fmovs %1,%0" : "=f" (r) : "f" (a));
   singletoop(res,r);
}

FLOATFUNC(FMOVD)
{ 
#ifdef SPARCV9
   UINT64 a = optodouble(op1), r;
   asm("fmovd %1,%0" : "=f" (r) : "f" (a));
   doubletoop(res,r);
#else
   v9inst("FMOVD");
#endif
}

FLOATFUNC(FNEGS)
{
   UINT32 a = optosingle(op1), r;
   asm("fnegs %1,%0" : "=f" (r) : "f" (a));
   singletoop(res,r);
}

FLOATFUNC(FNEGD)
{
#ifdef SPARCV9
   UINT64 a = optodouble(op1), r;
   asm("fnegd %1,%0" : "=f" (r) : "f" (a));
   doubletoop(res,r);
#else
   v9inst("FNEGD");
#endif
}

FLOATFUNC(FITOS)
{
   UINT32 a = optosingle(op1), r;
   asm("fitos %1,%0" : "=f" (r) : "f" (a));
   singletoop(res,r);
}
 
FLOATFUNC(FITOD)
{
   UINT32 a = optosingle(op1);
   UINT64 r;
   asm("fitod %1,%0" : "=f" (r) : "f" (a));
   doubletoop(res,r);
}

FLOATFUNC(FXTOS)
{
#ifdef SPARCV9
   UINT64 a = optodouble(op1);
   UINT32 r;
   asm("fxtos %1,%0" : "=f" (r) : "f" (a));
   singletoop(res,r);
#else
   v9inst("FXTOS");
#endif
}

FLOATFUNC(FXTOD)
{
#ifdef SPARCV9
   UINT64 r, a = optodouble(op1);
   asm("fxtod %1,%0" : "=f" (r) : "f" (a));
   doubletoop(res,r);
#else
   v9inst("FXTOD");
#endif
}


/* getline(). We provide two versions, one for Linux (which provides getdelim())
 * and one for the rest. This is just like GNU libc's getline() so if you
 * have a real glibc (unlike libc5 which #ifdefs it out (!)) you can use that.
 * Yes, I read the libc source when I was writing getline()...
 * Since we use fgets() internally we don't handle nulls in the input
 * stream terribly well. I don't consider this a problem in this application.
 */

#ifdef __USE_GNU
ssize_t getline (char **lineptr, size_t *n, FILE *stream)
{
   return getdelim (lineptr, n, '\n', stream);
}
#else /* not __USE_GNU */

ssize_t getline (char **lineptr, size_t *n, FILE *stream)
{
   /* Read a line from stream. Store it (including '\n\0') in a buffer,
    * and store the buffer address in *lineptr. On entry, *lineptr
    * holds the address of an initial buffer of length *n. 
    * We may realloc() this and update the addr and n. If *lineptr=NULL
    * and *n = 0 then we allocate an initial buffer for the caller.
    * We return number of chars read including \n but not \0, or -1 for
    * EOF or malloc()/realloc() failure.
    */
   int len = 0;                                   /* total number of chars read */
   
   /* sanity check on inputs */
   if (!lineptr || !n || ferror(stream))
      return -1;
   
   if (*lineptr == NULL || *n == 0)
   {
      *n = 120;                                   /* sensible initial length */
      *lineptr = (char*) malloc(*n);
      if (*lineptr == NULL)
         return -1;
   }
   
   for (;;)
   {
      if (fgets(*lineptr + len, *n - len, stream) == NULL)
         return -1;                               /* error or EOF */
      len += strlen(*lineptr+len);                /* add on number of chars just read */
      if (*(*lineptr+len-1) == '\n')              /* success! */
         return len;
      
      /* Nope, better increment the buffer a bit. */
      *n *= 2;                                    /* generous extra chunk of space */
      *lineptr = (char*) realloc(*lineptr, *n);
      if (*lineptr == NULL)
         return -1;
   }
}

#endif /* not __USE_GNU */
