/* $Id: tadpole.c,v 1.1.2.8 1997/03/28 00:23:07 ecd Exp $
 * tadpole.c: Routines for PROM emulation on poor Tadpole Sparcbook 1's.
 *
 * Copyright (C) 1996  Eddie C. Dost  (ecd@skynet.be)
 */


#include <linux/kernel.h>
#include <linux/string.h>

#include <asm/tadpole.h>
#include <asm/idprom.h>
#include <asm/oplib.h>
#include <asm/openprom.h>

/* #define TADPOLE_PROM_DEBUG */

#ifdef TADPOLE_PROM_DEBUG
static int tadpole_prom_initialized = 0;
#endif

extern unsigned long tadpole_memory_size;

static int   tadpole_kbdin = 0;
static int   tadpole_vgaout = 1;
static char *tadpole_bootpath = "";
static char *tadpole_bootargs = "";

static int   stdin;
static int   stdout;

static void  tadpole_vga_putchar(char);
static void  tadpole_vga_init(void);

static int   tadpole_dev_write(int, char *, int);
static int   tadpole_dev_read(int, char *, int);


struct property {
	char *name;
	char *value;
	int length;
};

struct node {
	struct property  *properties;
	struct node	 *child;
	struct node	 *next;
};

/*
 * Node properties:
 *
 * first the root node's properties.
 */
static unsigned char tadpole_idprom[32];
static unsigned long mbus_clk = 25000000;

static struct property root_properties[] = {
	{ "name", "SPARCbook", sizeof("SPARCbook") },
	{ "banner-name", "Tadpole SPARCbook 1", sizeof("Tadpole SPARCbook 1") },
	{ "compatible", "sun4m", sizeof("sun4m") },
	{ "idprom", tadpole_idprom, sizeof(tadpole_idprom) },
	{ "clock-frequency", (char *)&mbus_clk, sizeof(mbus_clk) },
	{ 0, 0, -1 }
};

/*
 * 'CY7C604A' properties:
 */
static unsigned long cy_line_size = 0x20;
static unsigned long cy_nlines = 0x800;
static unsigned long cy_nctx = 0x1000;
static unsigned long cy_mid = 0x8;

static struct property cypress_properties[] = {
	{ "name", "CY7C604A", sizeof("CY7C604A") },
	{ "device_type", "cpu", sizeof("cpu") },
	{ "cache-line-size", (char *)&cy_line_size, sizeof(cy_line_size) },
	{ "cache-nlines", (char *)&cy_nlines, sizeof(cy_nlines) },
	{ "mmu-nctx", (char *)&cy_nctx, sizeof(cy_nctx) },
	{ "mid", (char *)&cy_mid, sizeof(cy_mid) },
	{ 0, 0, -1 }
};

/*
 * 'obio' properties:
 */
static unsigned long obio_ranges[] = {
	0x2, 0x00000000, 0x1, 0x0c200000, 0x00200000,
	0x4, 0x00000000, 0x1, 0x0c400000, 0x00200000,
	0x6, 0x00000000, 0x1, 0x0c600000, 0x00200000,
	0xd, 0x00000000, 0x1, 0x0d000000, 0x00000100
};

static struct property obio_properties[] = {
	{ "name", "obio", sizeof("obio") },
	{ "device_type", "hierarchical", sizeof("hierarchical") },
	{ "ranges", (char *)obio_ranges, sizeof(obio_ranges) },
	{ 0, 0, -1 }
};

/*
 * 'eeprom' properties:
 */
static unsigned long eeprom_regs[] = {
	0x6, 0x00000000, 0x00000800
};

static struct property eeprom_properties[] = {
	{ "name", "eeprom", sizeof("eeprom") },
	{ "model", "mk48t02", sizeof("mk48t02") },
	{ "reg", (char *)eeprom_regs, sizeof(eeprom_regs) },
	{ 0, 0, -1 }
};


/*
 * 'pbus' properties:
 */
static unsigned long pbus_ranges[] = {
	0x0, 0x00000000, 0x1, 0x0c000000, 0x00200000,
	0x8, 0x00000000, 0x1, 0x0c800000, 0x00400000,
	0xc, 0x00000000, 0x1, 0x0cc00000, 0x00100000,
};
static unsigned long pbus_clk = 8000000;

static struct property pbus_properties[] = {
	{ "name", "pbus", sizeof("pbus") },
	{ "device_type", "hierarchical", sizeof("hierarchical") },
	{ "ranges", (char *)pbus_ranges, sizeof(pbus_ranges) },
	{ "clock-frequency", (char *)&pbus_clk, sizeof(pbus_clk) },
	{ 0, 0, -1 }
};

/*
 * 'mpi' properties:
 */
static unsigned long mpi_reg[] = {
	0xd, 0x00000000, 0x00000100
};

static unsigned long mpi_intr[] = {
	0xb, 0x0000001b
};

static struct property mpi_properties[] = {
	{ "name", "mpi", sizeof("mpi") },
	{ "reg", (char *)mpi_reg, sizeof(mpi_reg) },
	{ "intr", (char *)mpi_intr, sizeof(mpi_intr) },
	{ 0, 0, -1 }
};

/*
 * 'memory' properties:
 */
static unsigned long mem_reg[] = {
	0x0, 0x00000000, 0x00800000
};

static unsigned long mem_avail[] = {
	0x0, 0x00000000, 0x00800000
};

static struct property memory_properties[] = {
	{ "name", "memory", sizeof("memory") },
	{ "reg", (char *)mem_reg, sizeof(mem_reg) },
	{ "available", (char *)mem_avail, sizeof(mem_avail) },
	{ 0, 0, -1 }
};

/*
 * 'virtual-memory' properties:
 */
static unsigned long virt_mem_reg[] = {
	0x0, 0x00000000, 0x80000000,
	0x0, 0x80000000, 0x80000000
};

static unsigned long virt_mem_avail[] = {
	0x0, 0x00000000, 0x00000000
};

static struct property virtual_memory_properties[] = {
	{ "name", "virtual-memory", sizeof("virtual-memory") },
	{ "reg", (char *)virt_mem_reg, sizeof(virt_mem_reg) },
	{ "available", (char *)virt_mem_avail, sizeof(virt_mem_avail) },
	{ 0, 0, -1 }
};

/*
 * 'cgVGA' properties:
 */
static unsigned long vga_reg[] = {
	0xc, 0x000003b4, 0x0000002c,
	0x0, 0x00020000, 0x00040000
};

static unsigned long vga_intr[] = {
	0x00000006, 0x00000016
};

static unsigned long vga_width = 640;
static unsigned long vga_height = 480;
static unsigned long vga_depth = 4;
static unsigned long vga_linebytes = 80;

static struct property vga_properties[] = {
	{ "name", "cgVGA", sizeof("cgVGA") },
	{ "device_type", "display", sizeof("display") },
	{ "reg", (char *)vga_reg, sizeof(vga_reg) },
	{ "intr", (char *)vga_intr, sizeof(vga_intr) },
	{ "height", (char *)&vga_height, sizeof(vga_height) },
	{ "width", (char *)&vga_width, sizeof(vga_width) },
	{ "depth", (char *)&vga_depth, sizeof(vga_depth) },
	{ "linebytes", (char *)&vga_linebytes, sizeof(vga_linebytes) },
	{ "type", 0, 0 },
	{ 0, 0, -1 }
};

/*
 * 'nice' properties:
 */
static unsigned long nice_reg[] = {
	0x2, 0x00000000, 0x00000010
};

static unsigned long nice_intr[] = {
	0x00000007, 0x00000017
};

static struct property nice_properties[] = {
	{ "name", "nice", sizeof("nice") },
	{ "device_type", "network", sizeof("network") },
	{ "reg", (char *)nice_reg, sizeof(nice_reg) },
	{ "intr", (char *)nice_intr, sizeof(nice_intr) },
	{ 0, 0, -1 }
};

/*
 * 'id' properties:
 */
static unsigned long id_reg[] = {
	0x8, 0x000001f0, 0x00000002,
	0xc, 0x000001f0, 0x00000208,
};

static unsigned long id_intr[] = {
	0x00000005, 0x00000015
};

static struct property id_properties[] = {
	{ "name", "id", sizeof("id") },
	{ "device_type", "block", sizeof("block") },
	{ "reg", (char *)id_reg, sizeof(id_reg) },
	{ "intr", (char *)id_intr, sizeof(id_intr) },
	{ 0, 0, -1 }
};

/*
 * 'bpp' properties:
 */
static unsigned long bpp_reg[] = {
	0xc, 0x00000330, 0x00000003
};

static unsigned long bpp_intr[] = {
	0x00000001, 0x00000011
};

static struct property bpp_properties[] = {
	{ "name", "bpp", sizeof("bpp") },
	{ "device_type", "parallel", sizeof("parallel") },
	{ "reg", (char *)bpp_reg, sizeof(bpp_reg) },
	{ "intr", (char *)bpp_intr, sizeof(bpp_intr) },
	{ 0, 0, -1 }
};

/*
 * 'ncr' properties:
 */
static unsigned long ncr_reg[] = {
	0xc, 0x00000330, 0x00000003
};

static unsigned long ncr_intr[] = {
	0x00000001, 0x00000011
};

static struct property ncr_properties[] = {
	{ "name", "ncr", sizeof("ncr") },
	{ "device_type", "scsi", sizeof("scsi") },
	{ "reg", (char *)ncr_reg, sizeof(ncr_reg) },
	{ "intr", (char *)ncr_intr, sizeof(ncr_intr) },
	{ 0, 0, -1 }
};

/*
 * 'modem' properties:
 */
static unsigned long modem_reg[] = {
	0x4, 0x00000000, 0x00000008
};

static unsigned long modem_intr[] = {
	0x00000008, 0x00000018
};

static struct property modem_properties[] = {
	{ "name", "modem", sizeof("modem") },
	{ "device_type", "modem", sizeof("modem") },
	{ "reg", (char *)modem_reg, sizeof(modem_reg) },
	{ "intr", (char *)modem_intr, sizeof(modem_intr) },
	{ 0, 0, -1 }
};

/*
 * 'kbd' properties:
 */
static unsigned long kbd_intr[] = {
	0x0000000b, 0x0000001b
};

static struct property kbd_properties[] = {
	{ "name", "kbd", sizeof("kbd") },
	{ "device_type", "keyboard", sizeof("keyboard") },
	{ "intr", (char *)kbd_intr, sizeof(kbd_intr) },
	{ "keyboard", 0, 0 },
	{ 0, 0, -1 }
};

/*
 * 'mouse' properties:
 */
static unsigned long mouse_reg[] = {
	0xc, 0x00000310, 0x00000002
};

static unsigned long mouse_intr[] = {
	0x0000000a, 0x0000001a
};

static struct property mouse_properties[] = {
	{ "name", "mouse", sizeof("mouse") },
	{ "device_type", "mouse", sizeof("mouse") },
	{ "reg", (char *)mouse_reg, sizeof(mouse_reg) },
	{ "intr", (char *)mouse_intr, sizeof(mouse_intr) },
	{ 0, 0, -1 }
};

/*
 * 'upc' properties:
 */
static unsigned long upc_reg[] = {
	0xc, 0x000002fa, 0x00000001,
	0xc, 0x00000390, 0x00000002,
	0xc, 0x000003fa, 0x00000001
};

static struct property upc_properties[] = {
	{ "name", "upc", sizeof("upc") },
	{ "reg", (char *)upc_reg, sizeof(upc_reg) },
	{ 0, 0, -1 }
};

/*
 * 'urt' properties:
 */
static unsigned long urt_reg[] = {
	0xc, 0x000003f8, 0x00000008
};

static unsigned long urt_intr[] = {
	0x00000009, 0x00000019
};

static struct property urt_properties[] = {
	{ "name", "urt", sizeof("urt") },
	{ "device_type", "serial", sizeof("serial") },
	{ "reg", (char *)urt_reg, sizeof(urt_reg) },
	{ "intr", (char *)urt_intr, sizeof(urt_intr) },
	{ 0, 0, -1 }
};

/*
 * 'fd' properties:
 */
static unsigned long fd_reg[] = {
	0xc, 0x000003f0, 0x00000008
};

static unsigned long fd_intr[] = {
	0x00000002, 0x00000012
};

static char fd_status[] = "disabled";

static struct property fd_properties[] = {
	{ "name", "fd", sizeof("fd") },
	{ "device_type", "block", sizeof("block") },
	{ "reg", (char *)fd_reg, sizeof(fd_reg) },
	{ "intr", (char *)fd_intr, sizeof(fd_intr) },
	{ "status", fd_status, sizeof(fd_status) },
	{ 0, 0, -1 }
};


/*
 * 'options' properties:
 */
static struct property options_properties[] = {
	{ "name", "options", sizeof("options") },
	{ 0, 0, -1 }
};


/*
 * PROM node tree:
 */
static struct node options_node = {
		options_properties, 0, 0
};

static struct node virtual_memory_node = {
		virtual_memory_properties, 0, &options_node
};

static struct node memory_node = {
		memory_properties, 0, &virtual_memory_node
};

static struct node modem_node = {
			modem_properties, 0, 0
};

static struct node eeprom_node = {
			eeprom_properties, 0, &modem_node
};

static struct node nice_node = {
			nice_properties, 0, &eeprom_node
};

static struct node kbd_node = {
				kbd_properties, 0, 0
};

static struct node mpi_node = {
			mpi_properties, &kbd_node, &nice_node
};

static struct node obio_node = {
		obio_properties, &mpi_node, &memory_node
};

static struct node upc_node = {
			upc_properties, 0, 0
};

static struct node mouse_node = {
			mouse_properties, 0, &upc_node
};

static struct node urt_node = {
			urt_properties, 0, &mouse_node
};

static struct node ncr_node = {
			ncr_properties, 0, &urt_node
};

static struct node bpp_node = {
			bpp_properties, 0, &ncr_node
};

static struct node fd_node = {
			fd_properties, 0, &bpp_node
};

static struct node id_node = {
			id_properties, 0, &fd_node
};

static struct node vga_node = {
			vga_properties, 0, &id_node
};

static struct node pbus_node = {
		pbus_properties, &vga_node, &obio_node
};

static struct node cypress_node = {
		cypress_properties, 0, &pbus_node
};

static struct node root_node = {
	root_properties, &cypress_node, 0
};


/*
 * PROM node operations:
 */
static struct property *find_property(int node, char *name)
{
	struct property *prop;

	if (!node)
		return 0;

	prop = &((struct node *)node)->properties[0];
	if (!name || !strlen(name))
		return prop;

	while (prop && prop->name) {
		if (!strcmp(prop->name, name))
			return prop;
		prop++;
	}

	return 0;
}

static int
tadpole_no_nextnode(int node)
{
#ifdef TADPOLE_PROM_DEBUG
	struct property *prop;
#endif

	if (!node)
		return (int)&root_node;

	if (((struct node *)node)->next)
		return (int)((struct node *)node)->next;

#ifdef TADPOLE_PROM_DEBUG
	if (!tadpole_prom_initialized)
		return -1;

	prop = find_property(node, "name");
	if (prop)
		prom_printf("tadpole_nextnode(%08x[%s]) failed\n",
			    node, prop->value);
	else
		prom_printf("tadpole_nextnode(%08x) failed\n", node);
#endif
	return -1;
}

static int
tadpole_no_child(int node)
{
#ifdef TADPOLE_PROM_DEBUG
	struct property *prop;
#endif

	if (!node)
		return -1;

	if (((struct node *)node)->child)
		return (int)((struct node *)node)->child;

#ifdef TADPOLE_PROM_DEBUG
	if (!tadpole_prom_initialized)
		return -1;

	prop = find_property(node, "name");
	if (prop)
		prom_printf("tadpole_child(%08x[%s]) failed\n",
			    node, prop->value);
	else
		prom_printf("tadpole_child(%08x) failed\n", node);
#endif
	return -1;
}

static int
tadpole_no_proplen(int node, char *name)
{
	struct property *prop = find_property(node, name);

	if (prop)
		return prop->length;

#ifdef TADPOLE_PROM_DEBUG
	if (!tadpole_prom_initialized)
		return -1;

	prop = find_property(node, "name");
	if (prop)
		prom_printf("tadpole_proplen(%08x[%s], %s) failed\n",
			    node, prop->value, name);
	else
		prom_printf("tadpole_proplen(%08x, %s) failed\n", node, name);
#endif
	return -1;
}

static int
tadpole_no_getprop(int node, char *name, char *val)
{
	struct property *prop = find_property(node, name);

	if (prop) {
		memcpy(val, prop->value, prop->length);
		return prop->length;
	}

#ifdef TADPOLE_PROM_DEBUG
	if (!tadpole_prom_initialized)
		return -1;

	prop = find_property(node, "name");
	if (prop)
		prom_printf("tadpole_getprop(%08x[%s], %s) failed\n",
			    node, prop->value, name);
	else
		prom_printf("tadpole_getprop(%08x, %s) failed\n", node, name);
#endif
	return -1;
}

static int
tadpole_no_setprop(int node, char *name, char *val, int len)
{
#ifdef TADPOLE_PROM_DEBUG
	struct property *prop;
#endif

#ifdef TADPOLE_PROM_DEBUG
	if (!tadpole_prom_initialized)
		return -1;

	prop = find_property(node, "name");
	if (prop)
		prom_printf("tadpole_setprop(%08x[%s], %s) failed\n",
			    node, prop->value, name);
	else
		prom_printf("tadpole_setprop(%08x, %s) failed\n", node, name);
#endif
	return -1;
}

static char *
tadpole_no_nextprop(int node, char *name)
{
	struct property *prop = find_property(node, name);

	if (prop)
		return prop[1].name;

#ifdef TADPOLE_PROM_DEBUG
	if (!tadpole_prom_initialized)
		return 0;

	prop = find_property(node, "name");
	if (prop)
		prom_printf("tadpole_nextprop(%08x[%s], %s) failed\n",
			    node, prop->value, name);
	else
		prom_printf("tadpole_nextprop(%08x, %s) failed\n", node, name);
#endif
	return 0;
}


static struct linux_nodeops tadpole_nodeops = {
	tadpole_no_nextnode,
	tadpole_no_child,
	tadpole_no_proplen,
	tadpole_no_getprop,
	tadpole_no_setprop,
	tadpole_no_nextprop
};


static int
tadpole_getchar(void)
{
	char ch;
	if (tadpole_dev_read(stdin, &ch, 1) == 1)
		return ch;
	return -1;
}

static void
tadpole_putchar(int ch)
{
	char out = ch;
	tadpole_dev_write(stdout, &out, 1);
}

static int
tadpole_nbgetchar(void)
{
	char ch;
	if (tadpole_dev_read(stdin, &ch, 1) == 1)
		return ch;
	return -1;
}

static int
tadpole_nbputchar(int ch)
{
	char out = ch;
	if (tadpole_dev_write(stdout, &out, 1) == 1)
		return 0;
	return -1;
}

static void
tadpole_putstr(char *str, int len)
{
	tadpole_dev_write(stdout, str, len);
}

static void
tadpole_reboot(char *bootstr)
{
	__asm__ __volatile__ ("stba     %0, [%1] %2\n\t" : :
			      "r" (0x1c), "r" (0x0d0000a0), "i" (0x21));
	/* Assert the watchdog timer, until we know better ;-) */
	__asm__ __volatile__ ("stba     %0, [%1] %2\n\t" : :
			      "r" (0x40), "r" (0x0d000090), "i" (0x21));
	while (1) ;
}

static char ppbuf[1024];

static void
tadpole_printf(__const__ char *fmt, ...)
{
	va_list args;
	char ch, *bptr;
	int i;

	va_start(args, fmt);
	i = vsprintf(ppbuf, fmt, args);
	bptr = ppbuf;
	while ((ch = *(bptr++)) != 0) {
		if (ch == '\n')
			tadpole_putchar('\r');
		tadpole_putchar(ch);
	}
	va_end(args);
}

static void
tadpole_abort(void)
{
	while (1) {
		__asm__ __volatile__ ("stba     %0, [%1] %2\n\t" : :
				      "r" (0x02), "r" (0x0d0000a8), "i" (0x21));
	}
}

static void
tadpole_halt(void)
{
	while (1) {
		__asm__ __volatile__ ("stba     %0, [%1] %2\n\t" : :
				      "r" (0x02), "r" (0x0d0000a8), "i" (0x21));
	}
}

static void
tadpole_eval(char *str)
{
	if (!strcmp(str, "reset")) {
		__asm__ __volatile__ ("stba     %0, [%1] %2\n\t" : :
				      "r" (0x40), "r" (0x0d000090),
				      "i" (0x21));
		while (1) ;
	}
}

static unsigned int
tadpole_enaddr(int d, char *enaddr)
{
	unsigned char ack;
	int ok, try;

	memset(enaddr, 0, 6);
	enaddr[2] = 0x83;

	for (try = 0; try < 5; try++) {
		while (tadpole_mpi_serial_read_delay(&ack, 1)) {
			prom_printf("ack = %02x\n", ack);
			ack = 0;
		}
		tadpole_mpi_serial_write(0x10);
		ok = tadpole_mpi_serial_read_delay(&ack, 40000);
		ok &= tadpole_mpi_serial_read_delay(&enaddr[3], 40000);
		ok &= tadpole_mpi_serial_read_delay(&enaddr[4], 40000);
		ok &= tadpole_mpi_serial_read_delay(&enaddr[5], 40000);
		if (ok && (ack == 0xfe))
			break;
	}

	if (try == 5) {
#if 1
		prom_printf("tadpole_enaddr: read failed\n");
		prom_halt();
#endif
	}

	return 0;
}


static int
tadpole_inst2pkg(int d)
{
#ifdef TADPOLE_PROM_DEBUG
	if (tadpole_prom_initialized)
		prom_printf("tadpole_inst2pkg(%d) ", d);
#endif
	if (d == tadpole_kbdin) {
#ifdef TADPOLE_PROM_DEBUG
		if (tadpole_prom_initialized)
			prom_printf("keyboard\n");
#endif
		return (int)&kbd_node;
	} else if (d == tadpole_vgaout) {
#ifdef TADPOLE_PROM_DEBUG
		if (tadpole_prom_initialized)
			prom_printf("display\n");
#endif
		return (int)&vga_node;
	}
#ifdef TADPOLE_PROM_DEBUG
	if (tadpole_prom_initialized)
		prom_printf("none\n");
#endif
	return -1;
}

static char *
tadpole_dumb_mem_alloc(char *va, unsigned sz)
{
	return 0;
}

static void
tadpole_dumb_mem_free(char *va, unsigned sz)
{
}

static char *
tadpole_dumb_mmap(char *va, int bus, unsigned pa, unsigned sz)
{
	return 0;
}

static void
tadpole_dumb_munmap(char *va, unsigned sz)
{
}

static int
tadpole_dev_open(char *devpath)
{
	return -1;
}

static void
tadpole_dev_close(int d)
{
}

static int
tadpole_dev_read(int d, char *buf, int nbytes)
{
	return -1;
}

static int
tadpole_dev_write(int d, char *buf, int nbytes)
{
	int i;

	if (d == tadpole_vgaout) {
		for (i = 0; i < nbytes; i++)
			tadpole_vga_putchar(*buf++);
		return nbytes;
	}
	return -1;
}

static int
tadpole_dev_seek(int d, int hi, int lo)
{
	return -1;
}

static void tadpole_nop(void)
{
}


struct linux_romvec tadpole_romvec = {
	0,					/* pv_magic_cookie	*/
	37,					/* pv_romvers		*/
	0,					/* pv_plugin_revision	*/
	0,					/* pv_printrev		*/
	{
		0,				/* v0_totphys		*/
		0,				/* v0_prommap		*/
		0				/* v0_available		*/
	},
	&tadpole_nodeops,
	0,					/* pv_bootstr		*/
	{
		0,				/* v0_devopen		*/
		0,				/* v0_devclose		*/
		0,				/* v0_rdblkdev		*/
		0,				/* v0_wrblkdev		*/
		0,				/* v0_rdnetdev		*/
		0,				/* v0_wrnetdev		*/
		0,				/* v0_rdchardev		*/
		0,				/* v0_wrchardev		*/
		0				/* v0_seekdev		*/
	},
	0,					/* pv_stdin		*/
	0,					/* pv_stdout		*/
	tadpole_getchar,			/* pv_getchar		*/
	tadpole_putchar,			/* pv_putchar		*/
	tadpole_nbgetchar,			/* pv_nbgetchar		*/
	tadpole_nbputchar,			/* pv_nbputchar		*/
	tadpole_putstr,				/* pv_putstr		*/
	tadpole_reboot,				/* pv_reboot		*/
	tadpole_printf,				/* pv_printf		*/
	tadpole_abort,				/* pv_abort		*/
	0,					/* pv_ticks		*/
	tadpole_halt,				/* pv_halt		*/
	0,					/* pv_synchook		*/
	{
		0				/* v2_eval		*/
	},
	0,					/* pv_v0bootargs	*/
	tadpole_enaddr,				/* pv_enaddr		*/
	{
		&tadpole_bootpath,		/* v2_bootpath		*/
		&tadpole_bootargs,		/* v2_bootargs		*/
		&tadpole_kbdin,			/* v2_fd_stdin		*/
		&tadpole_vgaout			/* v2_fd_stdout		*/
	},
	{
		tadpole_inst2pkg,		/* v2_inst2pkg		*/
		tadpole_dumb_mem_alloc,		/* v2_dumb_mem_alloc	*/
		tadpole_dumb_mem_free,		/* v2_dumb_mem_free	*/
		tadpole_dumb_mmap,		/* v2_dumb_mmap		*/
		tadpole_dumb_munmap,		/* v2_dumb_munmap	*/
		tadpole_dev_open,		/* v2_dev_open		*/
		tadpole_dev_close,		/* v2_dev_close		*/
		tadpole_dev_read,		/* v2_dev_read		*/
		tadpole_dev_write,		/* v2_dev_write		*/
		tadpole_dev_seek,		/* v2_dev_seek		*/
		tadpole_nop,			/* v2_whee2		*/
		tadpole_nop,			/* v2_whee3		*/
	},
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	0,					/* pv_setctxt		*/
	0,					/* pv_cpustart		*/
	0,					/* pv_cpustop		*/
	0,					/* pv_cpuidle		*/
	0					/* pv_cpuresume		*/
};


void
tadpole_idprom_init(void)
{
	struct idprom *idprom = (struct idprom *)&tadpole_idprom;
	int i;

	memset(idprom, 0, sizeof(*idprom));

	idprom->id_format = 0x01;
	idprom->id_machtype = 0x81;
	tadpole_enaddr(0, idprom->id_ethaddr);

	idprom->id_sernum = (idprom->id_ethaddr[4] << 16) |
			    (idprom->id_ethaddr[5] << 8) |
			     idprom->id_ethaddr[6];

	idprom->id_cksum = 0;
	for (i = 0; i < 15; i++)
		idprom->id_cksum ^= tadpole_idprom[i];
}


void
tadpole_prom_init(void)
{
	tadpole_romvec.pv_romvers = 3;
	tadpole_romvec.pv_ticks = 0;
	tadpole_romvec.pv_fortheval.v2_eval = tadpole_eval;

	/* XXX: setup bootpath and bootargs. */

	stdin = tadpole_kbdin;
	stdout = tadpole_vgaout;

	/* Initialize memory reg property from boot argument in %o2. */
	mem_reg[2] = mem_avail[2] = tadpole_memory_size;

	tadpole_mpi_initialize();
	tadpole_vga_init();
	tadpole_idprom_init();

#ifdef TADPOLE_PROM_DEBUG
	tadpole_prom_initialized = 1;
#endif
}


#define READ_VGA_REG(reg)						\
({									\
	unsigned char val;						\
	__asm__ __volatile__ ("lduba	[%1] %2, %0\n\t" :		\
			      "=r" (val) :				\
			      "r" (0x0cc00000 | reg), "i" (0x21));	\
	val;								\
})

#define WRITE_VGA_REG(reg, val)						\
	__asm__ __volatile__ ("stba	%0, [%1] %2\n\t" : :		\
			      "r" (val), "r" (0x0cc00000 | reg),	\
			      "i" (0x21))

#define VGA_DELAY()							\
({									\
	int i;								\
	for (i = 0; i < 10; i++)					\
		/* nothing */;						\
})


void tadpole_scroll_up(void)
{
	unsigned long src, dst;
	int i;

	dst = 0x0c020000;
	src = dst + 16 * 80;

	/* Copy screen memory. */
	WRITE_VGA_REG(0x3ce, 0x05);
	WRITE_VGA_REG(0x3cf, 0x01);
	for (i = 0; i < 80 * 464; i++, src++, dst++) {
		__asm__ __volatile__ ("lduba	[%0] %2, %%g1\n\t"
				      "stba	%%g1, [%1] %2\n\t" : :
				      "r" (src), "r" (dst),
				      "i" (0x21) : "g1");
	}
	WRITE_VGA_REG(0x3cf, 0x00);

	/* Clear last line. */
	WRITE_VGA_REG(0x3ce, 0x00);
	WRITE_VGA_REG(0x3cf, 0x00);
	WRITE_VGA_REG(0x3ce, 0x01);
	WRITE_VGA_REG(0x3cf, 0x0f);
        for (; i < 80 * 480; i++) {
		__asm__ __volatile__ ("stba	%0, [%1] %2\n\t" : :
				      "r" (0x00),
				      "r" (0x0c020000 + i),
				      "i" (0x21));
        }
	WRITE_VGA_REG(0x3cf, 0x00);
}


void tadpole_vga_putchar(char c)
{
	unsigned long dst;
	unsigned long offset;
	unsigned char *font_row;
	static int esc_seen = 0;
	static int cmd_seen = 0;
	static int x = 0;
	static int y = 0;
	int i, skip;

	extern unsigned char vga_font[8192];

	if (cmd_seen) {
		switch (c) {
			case 'H':
				x = 0;
				y = 0;
				break;
			default:
				break;
		}
		cmd_seen = 0;
		return;
	}

	if (esc_seen) {
		switch (c) {
			case '[':
				cmd_seen = 1;
				break;
			default:
				break;
		}
		esc_seen = 0;
		return;
	}

	if (c == '\033') {
		esc_seen = 1;
		return;
	}

	if (c == '\r') {
		x = 0;
		return;
	}

	if (c == '\n') {
		y++;
		if (y > 29) {
			tadpole_scroll_up();
			y = 29;
		}
		return;
	}

	if (c == '\t') {
		skip = 8 - (x & 7);
		for (i = 0; i < skip; i++)
			tadpole_vga_putchar(' ');
		return;
	}

	font_row = &vga_font[c << 4];
	offset = 1280 * y + x;

	dst = 0x0c020000 + offset;

	WRITE_VGA_REG(0x3ce, 0x05);
	WRITE_VGA_REG(0x3cf, 0x02);
	WRITE_VGA_REG(0x3ce, 0x08);
	for (i = 0; i < 16; i++, font_row++, dst += 80) {
		WRITE_VGA_REG(0x3cf, *font_row);
		__asm__ __volatile__ ("lduba	[%1] %2, %%g0\n\t"
				      "stba	%0, [%1] %2\n\t" : :
				      "r" (0x01), "r" (dst), "i" (0x21));
		WRITE_VGA_REG(0x3cf, ~(*font_row));
		__asm__ __volatile__ ("lduba	[%1] %2, %%g0\n\t"
				      "stba	%0, [%1] %2\n\t" : :
				      "r" (0x00), "r" (dst), "i" (0x21));
	}
	WRITE_VGA_REG(0x3cf, 0xff);
	WRITE_VGA_REG(0x3ce, 0x05);
	WRITE_VGA_REG(0x3cf, 0x00);

	x++;
	if (x > 79) {
		x = 0;
		y++;
		if (y > 29) {
			tadpole_scroll_up();
			y = 29;
		}
	}
}


#define PANEL_TFT	1
#define PANEL_COLOUR	2
#define PANEL_MONO	3

void tadpole_vga_init(void)
{
	unsigned char crt[] = {
		0x5f, 0x4f, 0x50, 0x82, 0x54, 0x80, 0x0b, 0x3e,
		0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0xea, 0x8c, 0xdf, 0x28, 0x00, 0xe7, 0x04, 0xe3,
		0xff
	};
	unsigned char ac[] = {
		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
		0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
		0x01, 0x01, 0x0f, 0x00, 0x00
	};
	unsigned char gc[] = {
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x05, 0x0f,
		0xff
	};
	unsigned char horz[] = {
		0x50, 0x2d, 0x52, 0x28, 0x54, 0x10, 0x56, 0x2b, 0x58, 0x80,
		0x51, 0x5f, 0x53, 0x50, 0x55, 0x02, 0x57, 0x55, 0x59, 0x81
	};
	unsigned char vert[] = {
		0x6a, 0x45, 0x6b, 0x0b, 0x6c, 0xea, 0x6d, 0xac, 0x6e, 0xe7,
		0x6f, 0x04,
		0x5a, 0x45, 0x5b, 0x0b, 0x5c, 0xb7, 0x5d, 0x09, 0x5e, 0x96,
		0x5f, 0x05,
		0x4a, 0x45, 0x4b, 0x0b, 0x4c, 0xb0, 0x4d, 0x09, 0x4e, 0x63,
		0x4f, 0x06
	};
	unsigned char tft[] = {
		0x6a, 0x45, 0x6b, 0x0b, 0x6c, 0xea, 0x6d, 0xac, 0x6e, 0xe7,
		0x6f, 0x04,
		0x5a, 0x25, 0x5b, 0xdf, 0x5c, 0xdc, 0x5d, 0x0e, 0x5e, 0xd6,
		0x5f, 0x09,
		0x4a, 0x25, 0x4b, 0xdf, 0x4c, 0xbc, 0x4d, 0x0e, 0x4e, 0xb6,
		0x4f, 0x09
	};
	unsigned char seq[] = {
		0x03, 0x01, 0x0f, 0x00, 0x06
	};
	unsigned char dac[] = {
		0x3f, 0x3f, 0x3f, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x2a, 0x00, 0x2a, 0x00,
		0x00, 0x2a, 0x2a, 0x2a, 0x00, 0x00,
		0x2a, 0x00, 0x2a, 0x2a, 0x15, 0x00,
		0x2a, 0x2a, 0x2a, 0x15, 0x15, 0x15,
		0x15, 0x15, 0x3f, 0x15, 0x3f, 0x15,
		0x15, 0x3f, 0x3f, 0x3f, 0x15, 0x15,
		0x3f, 0x15, 0x3f, 0x3f, 0x3f, 0x15
	};
	struct property *prop;
	unsigned char value;
	int type;
	int i, j;

	/* Extension On */
	WRITE_VGA_REG(0x3ce, 0x0a);
	WRITE_VGA_REG(0x3cf, 0xec);

	/* Determine Display Type */
	WRITE_VGA_REG(0x3ce, 0xdc);
	value = READ_VGA_REG(0x3cf);
	if ((value & 0x03) == 0x01)
		type = PANEL_TFT;
	else {
		WRITE_VGA_REG(0x3ce, 0x81);
		value = READ_VGA_REG(0x3cf);
		if ((value & 0x08) || !(value & 0x01))
			type = PANEL_COLOUR;
		else
			type = PANEL_MONO;
	}

	for (prop = vga_properties; prop->name; prop++) {
		if (!strcmp(prop->name, "type")) {
			switch(type) {
				case PANEL_TFT:
					prop->value = "tft";
					prop->length = sizeof("tft");
					break;
				case PANEL_COLOUR:
					prop->value = "colour";
					prop->length = sizeof("colour");
					break;
				case PANEL_MONO:
					prop->value = "mono";
					prop->length = sizeof("mono");
					break;
				default:
					prop->name = 0;
					break;
			}
			break;
		}
	}


	/* Initialize General Parameters */
	for (i = 0; i < sizeof(horz); i += 2) {
		WRITE_VGA_REG(0x3ce, horz[i]);
		WRITE_VGA_REG(0x3cf, horz[i + 1]);
	}
	if (type == PANEL_TFT)
		for (i = 0; i < sizeof(tft); i += 2) {
			WRITE_VGA_REG(0x3ce, tft[i]);
			WRITE_VGA_REG(0x3cf, tft[i + 1]);
		}
	else
		for (i = 0; i < sizeof(vert); i += 2) {
			WRITE_VGA_REG(0x3ce, vert[i]);
			WRITE_VGA_REG(0x3cf, vert[i + 1]);
		}

	WRITE_VGA_REG(0x3ce, 0x80);
	WRITE_VGA_REG(0x3cf, 0x00);

	WRITE_VGA_REG(0x3ce, 0x81);
	value = READ_VGA_REG(0x3cf);
	if (type != PANEL_TFT)
		value |= 0x40;
	else
		value &= ~0x40;
	if (type == PANEL_MONO)
		value |= 0x81;
	WRITE_VGA_REG(0x3cf, value);

	WRITE_VGA_REG(0x3ce, 0x86);
	value = READ_VGA_REG(0x3cf);
	value |= 0x40;
	WRITE_VGA_REG(0x3cf, value);

	WRITE_VGA_REG(0x3ce, 0x83);
	value = READ_VGA_REG(0x3cf);
	if (type != PANEL_TFT)
		value |= 0x10;
	else
		value &= ~0x10;
	value |= 0x01;
	WRITE_VGA_REG(0x3cf, value);

	WRITE_VGA_REG(0x3ce, 0xd2);
	WRITE_VGA_REG(0x3cf, 0x2c);
	WRITE_VGA_REG(0x3ce, 0xda);
	value = READ_VGA_REG(0x3cf);
	value &= 0xf9;
	if (type != PANEL_TFT)
		value |= 0x04;
	WRITE_VGA_REG(0x3cf, value);

	WRITE_VGA_REG(0x3ce, 0xd8);
	WRITE_VGA_REG(0x3cf, 0x00);
	WRITE_VGA_REG(0x3ce, 0x92);
	WRITE_VGA_REG(0x3cf, 0x00);
	WRITE_VGA_REG(0x3ce, 0xc1);
	WRITE_VGA_REG(0x3cf, 0x00);

	/* Extension Off */
	WRITE_VGA_REG(0x3ce, 0x0a);
	WRITE_VGA_REG(0x3cf, 0xce);

	/* Write Misc. Output Register. */
	WRITE_VGA_REG(0x3c2, 0xe3);

	/* Sync Reset on. */
	WRITE_VGA_REG(0x3c4, 0x00);
	WRITE_VGA_REG(0x3c5, 0x01);

	/* Write Sequencer Registers. */
	WRITE_VGA_REG(0x3c4, 0x01);
	WRITE_VGA_REG(0x3c5, seq[1] | 0x20);
	for (i = 2; i < sizeof(seq); i++) {
		WRITE_VGA_REG(0x3c4, i);
		WRITE_VGA_REG(0x3c5, seq[i]);
	}

	/* Sync Reset off. */
	WRITE_VGA_REG(0x3c4, 0x00);
	WRITE_VGA_REG(0x3c5, 0x03);

	/* Unprotect CRT Registers 0-7. */
	WRITE_VGA_REG(0x3d4, 0x11);
	WRITE_VGA_REG(0x3d5, READ_VGA_REG(0x3d5) & 0x7f);
	/* Write CRT Registers. */
	for (i = 0; i < sizeof(crt); i++) {
		WRITE_VGA_REG(0x3d4, i);
		WRITE_VGA_REG(0x3d5, crt[i]);
	}

	/* Write Graphics Controller Registers. */
	for (i = 0; i < sizeof(gc); i++) {
		WRITE_VGA_REG(0x3ce, i);
		WRITE_VGA_REG(0x3cf, gc[i]);
	}

	/* Write Attribute Controller Registers. */
	for (i = 0; i < sizeof(ac); i++) {
		READ_VGA_REG(0x3da);
		VGA_DELAY();
		WRITE_VGA_REG(0x3c0, i);
		VGA_DELAY();
		WRITE_VGA_REG(0x3c0, ac[i]);
		VGA_DELAY();
	}

	/* Set Color Table. */
	WRITE_VGA_REG(0x3c6, 0xff);
	WRITE_VGA_REG(0x3c8, 0);
	for (i = 0; i < 16; i++) {
		for (j = 0; j < 16; j++) {
			WRITE_VGA_REG(0x3c9, dac[3 * j + 0]);
			WRITE_VGA_REG(0x3c9, dac[3 * j + 1]);
			WRITE_VGA_REG(0x3c9, dac[3 * j + 2]);
		}
	}

	/* Turn Screen on. */
	WRITE_VGA_REG(0x3c4, 0x01);
	WRITE_VGA_REG(0x3c5, READ_VGA_REG(0x3c5) & 0xdf);

	/* Enable Video Output. */
	READ_VGA_REG(0x3da);
	VGA_DELAY();
	WRITE_VGA_REG(0x3c0, 0x20);

	/* Clear Screen. */
	WRITE_VGA_REG(0x3ce, 0x00);
	WRITE_VGA_REG(0x3cf, 0x00);
	WRITE_VGA_REG(0x3ce, 0x01);
	WRITE_VGA_REG(0x3cf, 0x0f);
        for (i = 0; i < 80 * 480; i++) {
                __asm__ __volatile__ ("stba	%0, [%1] %2\n\t" : :
				      "r" (0x00), "r" (0x0c020000 + i),
				      "i" (0x21));
        }
	WRITE_VGA_REG(0x3cf, 0x00);
}
