/* $Id: tadpole_nice.c,v 1.1.2.11 1997/03/28 00:23:00 ecd Exp $
 * tadpole_nice.c: Tadpole SPARCbook 1 ethernet driver. The chip used
 *                 is a Fujitsu NICE (MB86960).
 *
 * Copyright (C) 1996  Eddie C. Dost  (ecd@skynet.be)
 */

static char *nice_name = "NICE";

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/tqueue.h>
#include <linux/in.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>

/* Used for the temporal inet entries and routing */
#include <linux/socket.h>
#include <linux/route.h>

#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/io.h>
#include <asm/byteorder.h>	/* Used by the checksum routines */
#include <asm/idprom.h>
#include <asm/sbus.h>
#include <asm/openprom.h>
#include <asm/oplib.h>


#undef DEBUG_DRIVER


#define CRC_POLYNOMIAL_LE 0x04c11db7UL
#define CRC_POLYNOMIAL_BE 0xedb88320UL


#define NICE_MAX_TX_PKTS	2	/* For each of the 2 buffers.
					 * I could go for more, but this
					 * only leads to more collisions
					 * on the net.
					 */
#define NICE_TX_BUFSIZ		4096	/* Enough for 2 packets. */


/*
 * TX status register:
 */
#define NICE_TSR_WR_BUSERR	0x01
#define NICE_TSR_16COL		0x02
#define NICE_TSR_COL		0x04
#define NICE_TSR_SHORT_PKT	0x10
#define NICE_TSR_RECV		0x20
#define NICE_TSR_NETBUSY	0x40
#define NICE_TSR_OK		0x80

#define NICE_TSR_CLEAR		0x87

/*
 * RX status register:
 */
#define NICE_RSR_OVERFLOW	0x01
#define NICE_RSR_CRC_ERROR	0x02
#define NICE_RSR_ALIGN_ERROR	0x04
#define NICE_RSR_SHORT_PKT	0x08
#define NICE_RSR_REMOTE_RESET	0x10
#define NICE_RSR_DMA_EOP	0x20
#define NICE_RSR_RD_BUSERR	0x40
#define NICE_RSR_PKT_READY	0x80

#define NICE_RSR_CLEAR		0xdf


/*
 * Transmit Mask Register
 */
#define NICE_TMR_WR_BUSERR	0x01
#define NICE_TMR_16COL		0x02
#define NICE_TMR_COL		0x04
#define NICE_TMR_OK		0x80

/*
 * Receive Mask Register
 */
#define NICE_RMR_OVERFLOW	0x01
#define NICE_RMR_CRC_ERROR	0x02
#define NICE_RMR_ALIGN_ERROR	0x04
#define NICE_RMR_SHORT_PKT	0x08
#define NICE_RMR_REMOTE_RESET	0x10
#define NICE_RMR_DMA_EOP	0x20
#define NICE_RMR_RD_BUSERR	0x40
#define NICE_RMR_PKT_READY	0x80

/*
 * Transmit Mode Register
 */
#define NICE_TMdR_DSC		0x01
#define NICE_TMdR_LBC		0x02
#define NICE_TMdR_TM		0x04
#define NICE_TMdR_COL_MASK	0xf0

/*
 * Receive Mode Register
 */
#define NICE_RMdR_AMM_MASK	0x03
#define NICE_RMdR_AMM_REJECT	0x00
#define NICE_RMdR_AMM_3BYTE	0x01
#define NICE_RMdR_AMM_6BYTEHASH	0x02
#define NICE_RMdR_AMM_RCVALL	0x03
#define NICE_RMdR_REMOTE_RESET	0x04
#define NICE_RMdR_SHORT_PKT	0x08
#define NICE_RMdR_SHORT_ADDR	0x10
#define NICE_RMdR_RCV_BAD_PKTS	0x20
#define NICE_RMdR_RCV_BUF_EMPTY	0x40
#define NICE_RMdR_TEST_CRC	0x80

/*
 * CSR register 0:
 */
#define NICE_CR0_MEMSIZ_MASK	0x03
#define NICE_CR0_MEMSIZ_8K	0x00
#define NICE_CR0_MEMSIZ_16K	0x01
#define NICE_CR0_MEMSIZ_32K	0x02
#define NICE_CR0_MEMSIZ_64K	0x03
#define NICE_CR0_TX_BUFSIZ_MASK	0x0c
#define NICE_CR0_TX_BUFSIZ_2K	0x00
#define NICE_CR0_TX_BUFSIZ_4K	0x04
#define NICE_CR0_TX_BUFSIZ_8K	0x08
#define NICE_CR0_TX_BUFSIZ_16K	0x0c
#define NICE_CR0_8BIT_SRAMBUS	0x10
#define NICE_CR0_8BIT_SYSBUS	0x20
#define NICE_CR0_100ns_SRAM	0x40
#define NICE_CR0_DISABLE	0x80

/*
 * CSR register 1:
 */
#define NICE_CR1_BIG_ENDIAN	0x01
#define NICE_CR1_EOP		0x02
#define NICE_CR1_BANK_MASK	0x0c
#define NICE_CR1_BANK_DLCR	0x00
#define NICE_CR1_BANK_HAST	0x04
#define NICE_CR1_BANK_BMR	0x08
#define NICE_CR1_RDY		0x10
#define NICE_CR1_NOT_STDBY	0x20
#define NICE_CR1_EEDCNTL_MASK	0xc0

/*
 * BMP Register:
 */
#define NICE_BMPR10_TMST	0x80
#define NICE_BMPR10_PKTS	0x7f

#define NICE_BMPR11_REPEAT_PKT	0x02
#define NICE_BMPR11_SKIP_PKT	0x03
#define NICE_BMPR11_CONT16_REP	0x06
#define NICE_BMPR11_CONT16_SKIP	0x07

#define NICE_BMPR12_DMA_WENABLE	0x01
#define NICE_BMPR12_DMA_RENABLE	0x02

#define NICE_BMPR13_BURST_1	0x00
#define NICE_BMPR13_BURST_4	0x01
#define NICE_BMPR13_BURST_8	0x02
#define NICE_BMPR13_BURST_12	0x03


#define set_bank(csr, bank)		\
({					\
	unsigned char csr_val;		\
	csr_val = csr;			\
	csr_val &= ~NICE_CR1_BANK_MASK;	\
	csr_val |= (bank << 2);		\
	csr = csr_val;			\
})

struct nice_regs {
	volatile unsigned char tx_status;
	volatile unsigned char rx_status;
	volatile unsigned char tx_int_mask;
	volatile unsigned char rx_int_mask;
	volatile unsigned char tx_mode;
	volatile unsigned char rx_mode;
	volatile unsigned char csr0;
	volatile unsigned char csr1;
	union {
		struct {
			volatile unsigned char  node_id[6];
			volatile unsigned short tdr_count;
		} rb0;
		struct {
			volatile unsigned char  mask_addr[8];
		} rb1;
		struct {
			volatile unsigned short bm_port;
			volatile unsigned char  tx_ctl0;
			volatile unsigned char  tx_ctl1;
			volatile unsigned char  dma_enable;
			volatile unsigned char  dma_burst_ctl;
			volatile unsigned char  rx_bptr;
			volatile unsigned char  reserved;
		} rb2;
	} bank;
};

struct nice_priv {
	struct nice_regs *nr;
	struct device *dev;
	unsigned int tx_packets;
	unsigned int queue_packets;
	long queue_free;
	unsigned int hard_col_pkt;
	unsigned int hard_collisions;
	unsigned long filter[2];
	struct enet_statistics stat;
};

static struct device *nice_dev;


static void nice_reset(struct nice_priv *np)
{
	struct nice_regs *nr = np->nr;
	int i;

	nr->csr0 = NICE_CR0_MEMSIZ_64K | NICE_CR0_TX_BUFSIZ_8K |
		   NICE_CR0_100ns_SRAM | NICE_CR0_DISABLE;
	nr->csr1 = NICE_CR1_NOT_STDBY;

	nr->tx_mode = NICE_TMdR_LBC;
	if (np->dev->flags & IFF_PROMISC) {
		nr->rx_mode = NICE_RMdR_AMM_RCVALL | NICE_RMdR_REMOTE_RESET;
	} else {
		nr->rx_mode = NICE_RMdR_AMM_6BYTEHASH | NICE_RMdR_REMOTE_RESET;
	}

	nr->tx_status = NICE_TSR_CLEAR;
	nr->rx_status = NICE_RSR_CLEAR;

	nr->tx_int_mask = NICE_TMR_WR_BUSERR | NICE_TMR_16COL |
			  NICE_TMR_COL | NICE_TMR_OK;
	nr->rx_int_mask = NICE_RMR_OVERFLOW | NICE_RMR_CRC_ERROR |
			  NICE_RMR_ALIGN_ERROR | NICE_RMR_SHORT_PKT |
			  NICE_RMR_RD_BUSERR | NICE_RMR_PKT_READY;

	set_bank(nr->csr1, 0);
	for (i = 0; i < ETH_ALEN; i++)
		nr->bank.rb0.node_id[i] = np->dev->dev_addr[i];

	set_bank(nr->csr1, 1);
	nr->bank.rb1.mask_addr[0] = np->filter[0] & 0xff;
	nr->bank.rb1.mask_addr[1] = (np->filter[0] >> 8) & 0xff;
	nr->bank.rb1.mask_addr[2] = (np->filter[0] >> 16) & 0xff;
	nr->bank.rb1.mask_addr[3] = (np->filter[0] >> 24) & 0xff;
	nr->bank.rb1.mask_addr[4] = np->filter[1] & 0xff;
	nr->bank.rb1.mask_addr[5] = (np->filter[2] >> 8) & 0xff;
	nr->bank.rb1.mask_addr[6] = (np->filter[3] >> 16) & 0xff;
	nr->bank.rb1.mask_addr[7] = (np->filter[4] >> 24) & 0xff;

	np->tx_packets = 0;
	np->queue_packets = 0;
	np->queue_free = NICE_TX_BUFSIZ;
	np->hard_col_pkt = 0;

	set_bank(nr->csr1, 2);
	nr->bank.rb2.tx_ctl1 = NICE_BMPR11_REPEAT_PKT;

	nr->csr0 &= ~(NICE_CR0_DISABLE);
}


static inline void nice_tx(int irq, struct nice_priv *np)
{
	struct nice_regs *nr = np->nr;
	unsigned char status;

	status = nr->tx_status;
	if (status & NICE_TSR_OK) {
		nr->tx_status = NICE_TSR_OK;
		np->stat.tx_packets += np->tx_packets;
		np->tx_packets = np->queue_packets;
		if (np->tx_packets) {
			np->queue_packets = 0;
			np->queue_free = NICE_TX_BUFSIZ;
			np->hard_col_pkt = 0;
#ifdef DEBUG_DRIVER
			if (np->tx_packets > 1)
				printk("%s: TX OK, next %d packets\n",
					np->dev->name, np->tx_packets);
#endif
			nr->bank.rb2.tx_ctl0 = NICE_BMPR10_TMST|np->tx_packets;
		}
	}
	if (status & NICE_TSR_WR_BUSERR) {
		nr->tx_status = NICE_TSR_WR_BUSERR;
		np->stat.tx_errors++;
		np->stat.tx_fifo_errors++;
#ifdef DEBUG_DRIVER
		printk("%s: Buffer Memory Write Error\n", np->dev->name);
#endif
	}
	if (status & NICE_TSR_COL) {
		nr->tx_status = NICE_TSR_COL;
		np->stat.collisions += (nr->tx_mode & NICE_TMdR_COL_MASK) >> 4;
#ifdef DEBUG_DRIVER
		printk("%s: COL\n", np->dev->name);
#endif
	}
	if (status & NICE_TSR_16COL) {
		unsigned char n;

		nr->tx_status = NICE_TSR_16COL;
		np->stat.collisions += 16;
		n = nr->bank.rb2.tx_ctl0 & NICE_BMPR10_PKTS;
		if (n == np->hard_col_pkt) {
			if (++np->hard_collisions > 2) {
				nr->bank.rb2.tx_ctl1 = NICE_BMPR11_SKIP_PKT;
#ifdef DEBUG_DRIVER
				printk("%s: skipping packet after %d COL\n",
				       np->dev->name, 16 * np->hard_collisions);
#endif
			} else {
				nr->bank.rb2.tx_ctl1 = NICE_BMPR11_REPEAT_PKT;
#ifdef DEBUG_DRIVER
				printk("%s: %d COL\n", np->dev->name,
				       16 * np->hard_collisions);
#endif
			}
		} else {
			np->hard_col_pkt = n;
			np->hard_collisions = 1;
			nr->bank.rb2.tx_ctl1 = NICE_BMPR11_REPEAT_PKT;
#ifdef DEBUG_DRIVER
			printk("%s: 16 COL\n", np->dev->name);
#endif
		}
	}
	if (status & NICE_TSR_SHORT_PKT) {
		np->stat.collisions++;
#ifdef DEBUG_DRIVER
		printk("%s: short pkt\n", np->dev->name);
#endif
	}
}


static inline void nice_read_buffer(struct nice_priv *np)
{
	struct nice_regs *nr = np->nr;
	struct sk_buff *skb;
	unsigned short len, stat;
	unsigned short *data;
	int i;
#ifdef DEBUG_DRIVER
	int packets = 0;
#endif

	while (!(nr->rx_mode & NICE_RMdR_RCV_BUF_EMPTY)) {
		stat = nr->bank.rb2.bm_port;
		len = nr->bank.rb2.bm_port;
		len = (len >> 8) | ((len & 0xff) << 8);
		skb = dev_alloc_skb(len + 2);
		if (!skb) {
			volatile unsigned short dummy;

			printk("%s: Memory squeeze, packet deferred\n",
			       np->dev->name);
			np->stat.rx_dropped++;
			for (i = 0; i < len; i += 2)
				dummy = nr->bank.rb2.bm_port;
			continue;
		}
		skb->dev = np->dev;
		skb_reserve(skb, 2);
		skb_put(skb, len);
		data = (unsigned short *)skb->data;
		for (i = 0; i < len; i += 2)
			*data++ = nr->bank.rb2.bm_port;

		skb->protocol = eth_type_trans(skb, np->dev);
		netif_rx(skb);
		np->stat.rx_packets++;
#ifdef DEBUG_DRIVER
		++packets;
		if (packets > 7) {
			printk("%s: RX %d packets\n",
				np->dev->name, packets);
		}
#endif
	}
}


static inline void nice_rx(int irq, struct nice_priv *np)
{
	struct nice_regs *nr = np->nr;
	unsigned char status;

	status = nr->rx_status;
	if (status & NICE_RSR_PKT_READY) {
		nr->rx_status = NICE_RSR_PKT_READY;
		nice_read_buffer(np);
	}

	if (status & NICE_RSR_OVERFLOW) {
		nr->rx_status = NICE_RSR_OVERFLOW;
		np->stat.rx_errors++;
		np->stat.rx_fifo_errors++;
#ifdef DEBUG_DRIVER
		printk("%s: RX Overflow\n", np->dev->name);
#endif
		nice_read_buffer(np);
		return;
	}
	if (status & NICE_RSR_CRC_ERROR) {
		nr->rx_status = NICE_RSR_CRC_ERROR;
		np->stat.rx_errors++;
		np->stat.rx_crc_errors++;
#ifdef DEBUG_DRIVER
		printk("%s: RX CRC Error\n", np->dev->name);
#endif
	}
	if (status & NICE_RSR_ALIGN_ERROR) {
		nr->rx_status = NICE_RSR_ALIGN_ERROR;
		np->stat.rx_errors++;
		np->stat.rx_over_errors++;
#ifdef DEBUG_DRIVER
		printk("%s: RX Align Error\n", np->dev->name);
#endif
	}
	if (status & NICE_RSR_SHORT_PKT) {
		nr->rx_status = NICE_RSR_SHORT_PKT;
		np->stat.rx_errors++;
		np->stat.rx_frame_errors++;
#ifdef DEBUG_DRIVER
		printk("%s: RX Short Packet, TX CTL: %02x\n",
			np->dev->name, nr->bank.rb2.tx_ctl0);
#endif
	}
	if (status & NICE_RSR_REMOTE_RESET) {
#ifdef DEBUG_DRIVER
		printk("%s: RX Remote Reset\n", np->dev->name);
#endif
		nice_reset(np);
		return;
	}
	if (status & NICE_RSR_RD_BUSERR) {
		nr->rx_status = NICE_RSR_RD_BUSERR;
		np->stat.rx_errors++;
		np->stat.rx_fifo_errors++;
#ifdef DEBUG_DRIVER
		printk("%s: Buffer Memory Read Error\n", np->dev->name);
#endif
	}
}


static void nice_interrupt(int irq, void *id, struct pt_regs *regs)
{
	struct device *dev = (struct device *)id;
	struct nice_priv *np = (struct nice_priv *)dev->priv;
	struct nice_regs *nr = np->nr;

	if (dev->interrupt)
		printk("%s: recursive interrupt\n", dev->name);
	dev->interrupt = 1;

	if (nr->tx_status)
		nice_tx(irq, np);

	if (nr->rx_status)
		nice_rx(irq, np);

	if ((np->queue_packets < NICE_MAX_TX_PKTS) && dev->tbusy) {
		dev->tbusy = 0;
		mark_bh(NET_BH);
	}

	dev->interrupt = 0;
}


static int nice_open(struct device *dev)
{
	struct nice_priv *np = (struct nice_priv *)dev->priv;

	if (request_irq(dev->irq, &nice_interrupt, SA_SHIRQ, nice_name,
			(void *)dev)) {
		printk("%s: can't get IRQ %d\n", dev->name, dev->irq);
		return -EAGAIN;
	}

	dev->tbusy = 0;
	dev->interrupt = 0;
	dev->start = 1;

	np->filter[0] = np->filter[1] = 0;

	nice_reset(np);
	return 0;
}

static int nice_close(struct device *dev)
{
	free_irq(dev->irq, (void *)dev);

	dev->tbusy = 1;
	dev->start = 0;

	return 0;
}

static int nice_start_xmit(struct sk_buff *skb, struct device *dev)
{
	struct nice_priv *np = (struct nice_priv *)dev->priv;
	struct nice_regs *nr = np->nr;
	unsigned long flags;
	unsigned short len, skblen;
	unsigned short *data;
	int i;

	if (dev->tbusy) {
		int ticks = jiffies - dev->trans_start;

		if (ticks < 100)
			return -1;
		else {
			printk("%s: TX timed out, status %02x, ctl %02x\n",
			       dev->name, nr->tx_status, nr->bank.rb2.tx_ctl0);
			np->stat.tx_errors++;
			dev->tbusy = 0;
			dev->trans_start = jiffies;
			nice_reset(np);
			dev_kfree_skb(skb, FREE_WRITE);
			return 0;
		}
	}

	if (!skb) {
		printk("%s: skb is NULL\n", dev->name);
		dev_tint(dev);
		return 0;
	}

	if (skb->len <= 0) {
		printk("%s: skb len is %ld\n", dev->name, skb->len);
		return 0;
	}

	if (set_bit (0, (void *) &dev->tbusy) != 0) {
		printk ("%s: Transmitter access conflict.\n", dev->name);
		return -1;
	}

	skblen = skb->len;
	len = (skblen <= ETH_ZLEN) ? ETH_ZLEN : ((skblen + 1) & ~1);

	save_and_cli(flags);
	if ((np->queue_free - (len + 2)) < 0) {
#ifdef DEBUG_DRIVER
		printk("%s: start_xmit: buffer, pkts %d, free %ld, len %d\n",
			dev->name, np->queue_packets, np->queue_free, len);
#endif
		restore_flags(flags);
		return -1;
	}

	nr->bank.rb2.bm_port = (len & 0xff) << 8 | (len >> 8);
	data = (unsigned short *)skb->data;
	for (i = 0; i < skblen; i += 2)
		nr->bank.rb2.bm_port = *data++;
	for ( ; i < len; i += 2)
		nr->bank.rb2.bm_port = 0;

	np->queue_free -= len + 2;
	np->queue_packets++;
	if (!np->tx_packets) {
		np->tx_packets = np->queue_packets;
		np->queue_packets = 0;
		np->queue_free = NICE_TX_BUFSIZ;
		np->hard_col_pkt = 0;
#ifdef DEBUG_DRIVER
		if (np->tx_packets > 1)
		printk("%s: start_xmit: start %d packets\n", dev->name,
			np->tx_packets);
#endif
		nr->bank.rb2.tx_ctl0 = NICE_BMPR10_TMST | np->tx_packets;
	}

	if (np->queue_packets < NICE_MAX_TX_PKTS)
		dev->tbusy = 0;

	dev->trans_start = jiffies;
	restore_flags(flags);

	dev_kfree_skb(skb, FREE_WRITE);
	return 0;
}

static struct enet_statistics *nice_get_stats(struct device *dev)
{
	struct nice_priv *np = (struct nice_priv *)dev->priv;

	return &np->stat;
}

static void nice_load_multicast(struct device *dev)
{
	struct nice_priv *np = (struct nice_priv *)dev->priv;
	struct nice_regs *nr = np->nr;
	u16 *mcast_table = (u16 *)&np->filter;
	struct dev_mc_list *dmi = dev->mc_list;
	char *addrs;
	int i, j, bit, byte;
	u32 crc, poly = CRC_POLYNOMIAL_LE;

	set_bank(nr->csr1, 1);
	if (dev->flags & IFF_ALLMULTI) {
		np->filter[0] = 0xffffffff;
		np->filter[1] = 0xffffffff;
		for (i = 0; i < 8; i++)
			nr->bank.rb1.mask_addr[i] = 0xff;
		set_bank(nr->csr1, 2);
		return;
	}

	np->filter[0] = 0;
	np->filter[1] = 0;

	for (i = 0; i < dev->mc_count; i++) {
		addrs = dmi->dmi_addr;
		dmi = dmi->next;

		if (!(*addrs & 1))
			continue;

		crc = 0xffffffff;
		for (byte = 0; byte < 6; byte++)
			for (bit = *addrs++, j = 0; j < 8; j++, bit >>= 1) {
				int test;

				test = ((bit ^ crc) & 0x01);
				crc >>= 1;
				if (test)
					crc = crc ^ poly;
			}
		crc = crc >> 26;
		mcast_table[crc >> 4] |= 1 << (crc & 0xf);
	}
	nr->bank.rb1.mask_addr[0] = np->filter[0] & 0xff;
	nr->bank.rb1.mask_addr[1] = (np->filter[0] >> 8) & 0xff;
	nr->bank.rb1.mask_addr[2] = (np->filter[0] >> 16) & 0xff;
	nr->bank.rb1.mask_addr[3] = (np->filter[0] >> 24) & 0xff;
	nr->bank.rb1.mask_addr[4] = np->filter[1] & 0xff;
	nr->bank.rb1.mask_addr[5] = (np->filter[2] >> 8) & 0xff;
	nr->bank.rb1.mask_addr[6] = (np->filter[3] >> 16) & 0xff;
	nr->bank.rb1.mask_addr[7] = (np->filter[4] >> 24) & 0xff;
	set_bank(nr->csr1, 2);
}

static void nice_set_multicast_list(struct device *dev)
{
	struct nice_priv *np = (struct nice_priv *)dev->priv;
	struct nice_regs *nr = np->nr;
#ifdef DEBUG_DRIVER
	int pr;

	pr = 0;
#endif
	while (dev->tbusy) {
#ifdef DEBUG_DRIVER
		if (!pr) {
			printk("nice_set_multi: tbusy\n");
			pr = 1;
		}
#endif
		schedule();
	}
	set_bit (0, (void *) &dev->tbusy);

	nr->csr0 |= NICE_CR0_DISABLE;
	if (dev->flags & IFF_PROMISC) {
		nr->rx_mode = NICE_RMdR_AMM_RCVALL | NICE_RMdR_REMOTE_RESET;
	} else {
		nr->rx_mode = NICE_RMdR_AMM_6BYTEHASH | NICE_RMdR_REMOTE_RESET;
		nice_load_multicast(dev);
	}
	nr->csr0 &= ~(NICE_CR0_DISABLE);
	dev->tbusy = 0;
}

static int nice_init(struct device *dev, int node)
{
	struct linux_prom_registers regs[1];
	struct linux_prom_irqs irqs[1];
	static int version_printed = 0;
	struct nice_priv *np;
	struct nice_regs *nr;
	int i;

	if (!dev)
		dev = init_etherdev(0, sizeof(struct nice_priv) + 7);
	else {
		dev->priv = kmalloc(sizeof(struct nice_priv) + 7, GFP_KERNEL);
		if (!dev->priv)
			return -ENOMEM;
	}

	if (!version_printed) {
		char *revision = "$Revision: 1.1.2.11 $", *r, *e;
		char *date = "$Date: 1997/03/28 00:23:00 $", *d;
		int j;

		if ((r = strchr(revision, ':')) &&
		    (e = strchr(r, '$')))
			r += 2, e--, i = e - r;
		else
			r = revision, i = -1;

		if ((d = strchr(date, ':')) && (d += 2) &&
		    (e = strchr(d, ' ')))
			j = e - d;
		else
			d = date, j = -1;

		printk ("%s: NICE v%.*s %.*s Eddie C. Dost (ecd@skynet.be)\n",
			dev->name, i, r, j, d);
		version_printed++;
	}

	nice_dev = dev;

	printk ("%s: NICE ", dev->name);

	/* Copy the IDPROM ethernet address to the device structure, later we
	 * will copy the address in the device structure to the nice CSR.
	 */
	for (i = 0; i < ETH_ALEN; i++)
		printk ("%2.2x%c", dev->dev_addr[i] = idprom->id_ethaddr[i],
			i == 5 ? ' ' : ':');
	printk ("\n");

	/* Get the IO region */
	prom_getproperty(node, "reg", (char *)regs, sizeof(regs));
	dev->base_addr = (long)regs[0].phys_addr;

        /* Map the registers read/write */
	prom_apply_obio_ranges(regs, 0x1);
        nr = sparc_alloc_io(regs[0].phys_addr, 0,
			     sizeof (struct nice_regs), nice_name,
			     regs[0].which_io, 0x0);

	/* Get the Interrupt. */
        prom_getproperty(node, "intr", (char *)irqs, sizeof(irqs));

	/* Make certain the data structures used by the NICE are aligned. */
	dev->priv = (void *)(((int)dev->priv + 7) & ~7);
	np = (struct nice_priv *) dev->priv;
	memset ((char *)dev->priv, 0, sizeof (struct nice_priv));

	np->nr = nr;
	np->dev = dev;

	dev->open = &nice_open;
	dev->stop = &nice_close;
	dev->hard_start_xmit = &nice_start_xmit;
	dev->get_stats = &nice_get_stats;
	dev->set_multicast_list = &nice_set_multicast_list;

	dev->irq = (unsigned char) irqs[0].pri;
	dev->dma = 0;
	ether_setup (dev);

	return 0;
}


/* Find all the lance cards on the system and initialize them */
int tadpole_nice_probe (struct device *dev)
{
	static int called = 0;
	int node;
    
	if(called)
		return ENODEV;
	called++;

	if (!((node = prom_getchild(prom_root_node)) &&
	      (node = prom_searchsiblings(node, "obio")) &&
	      (node = prom_getchild(node)) &&
	      (node = prom_searchsiblings(node, "nice"))))
		return ENODEV;

	return nice_init(dev, node);
}
